import os
import os
import re
import string
import time
import unicodedata

import pandas as pd
import rdflib
import yaml
from django.conf import settings
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import FieldError, MultipleObjectsReturned
from django.db.models.fields import CharField as TCharField
from django.db.models.fields import FloatField as TFloatField
from django.db.models.fields.related import ForeignKey as TForeignKey
from django.db.models.fields.related import ManyToManyField as TManyToMany
from rdflib import URIRef, RDFS
from rdflib.namespace import SKOS, OWL

from apis_core.apis_labels.models import Label
from apis_core.apis_metainfo.models import Uri as genUri, Collection, Uri
from apis_core.apis_vocabularies.models import LabelType
from apis_core.default_settings.RDF_settings_new import sameAs

APIS_RDF_YAML_SETTINGS = getattr(
        settings,
        'APIS_RDF_YAML_SETTINGS',
        os.path.join(os.path.dirname(__file__), '../default_settings', 'RDF_default_settings.yml')
    )

APIS_RDF_URI_SETTINGS = getattr(
        settings,
        'APIS_RDF_URI_SETTINGS',
        os.path.join(os.path.dirname(__file__), '../default_settings', "URI_replace_settings.yml")
    )


def clean_uri(sett, uri):
    for dom in sett['mappings']:
        if dom['domain'] in uri:
            m = re.match(dom['regex'], uri)
            if m:
                uri = dom['replace'].format(m.group(1))
    return uri


class PartialFormatter(string.Formatter):

    def __init__(self, missing='-', bad_fmt='!!'):
        self.missing, self.bad_fmt = missing, bad_fmt

    def get_field(self, field_name, args, kwargs):
        try:
            val = super().get_field(field_name, args, kwargs)
        except (KeyError, AttributeError):
            val = None, field_name
        return val

    def format_field(self, value, spec):
        if value is None:
            return self.missing
        try:
            return super(PartialFormatter, self).format_field(value, spec)
        except ValueError:
            if self.bad_fmt is not None:
                return self.bad_fmt
            else:
                raise


fmt = PartialFormatter()


class RDFParser(object):

    _reserved_uris = dict()

    @property
    def _settings_complete(self):
        sett = yaml.load(open(self._rdf_settings_file, 'r'))
        return sett

    @property
    def _settings(self):
        """
        Reads settings file and saves it
        :return: (dict) dict of settings file
        """
        sett = yaml.load(open(self._rdf_settings_file, 'r'))
        res = {'data': []}
        for v in sett[self.kind]['data']:
            if v['base_url'] in self.uri:
                res['data'] = v
        res['matching'] = sett[self.kind]['matching']
        res['sameAs'] = sameAs
        return res

    @staticmethod
    def _clean_uri_store():
        del_items = []
        for key, value in RDFParser._reserved_uris.items():
            if (time.time() - value) > (30 * 60):
                del_items.append(key)
        for d in del_items:
            del RDFParser._reserved_uris[d]

    def _exist(self, uri, uri_check=True):

        if self.objct.objects.filter(uri__uri=uri).count() > 0:
            return True, self.objct.objects.get(uri__uri=uri)
        else:
            if uri in RDFParser._reserved_uris.keys() and uri_check:
                if (time.time() - RDFParser._reserved_uris[uri]) < (self._preserve_uri_minutes * 60):
                    raise ValueError("URI used by other instance")
                else:
                    RDFParser._reserved_uris[uri] = time.time()
                    return False, False
            elif uri_check:
                RDFParser._reserved_uris[uri] = time.time()
                return False, False
            else:
                return False, False

    def _parse(self):
        """
        Parses the URI with the definition in the settings file
        :return: (dict) dict of Pandas DataFrames containing the variables obtained from SPARQL query
        """
        g = rdflib.Graph()
        g.parse(f"{self.uri}")
        if self._use_preferred:
            lst_pref = []
            pref_fin = []
            u = None
            count = 0
            u = self._settings_complete[self.kind]['data'][count]['base_url']
            while u:
                if u not in self.uri:
                    lst_pref.append(u)
                    count += 1
                    u = self._settings_complete[self.kind]['data'][count]['base_url']
                else:
                    u = False
            for o in g.objects(URIRef(self.uri), OWL.sameAs):
                for idx1, o2 in enumerate(lst_pref):
                    if o2 in str(o):
                        pref_fin.append((idx1, o))
            if len(pref_fin) > 0:
                pref_fin.sort(key=lambda tup: tup[0])
                self._parse = g.parse(pref_fin[0][1])
                self.uri = pref_fin[0][1]
        self._graph = g
        res = dict()
        for s in self._settings['data']['attributes']:
            sp = s.get('sparql', False)
            if sp:
                sp2 = self._sparql_to_pandas(g.query(sp, initBindings={'subject': URIRef(self.uri)}))
                res[s.get('name', 'no identifier provided')] = pd.DataFrame(sp2)
        self._attributes = res
        return res

    def _add_same_as(self):
        """
        adds the same as links as defined in the settings to the entity
        """
        for sa in self._settings['sameAs']:
            for su in self._graph.objects((self._subject, URIRef(sa))):
                Uri.objects.create(uri=su, entity=self.objct)

    def get_or_create(self, depth=2):
        """
        Gets or creates object with given URI and entity type.
        :param depth: (int) depth of related objects to be followed
        :return: (object) created object
        """
        if not self.created:
            return self.objct
        else:
            if not self.saved:
                self.create_objct(depth=depth)
                ob = self.save()
                return ob
            else:
                return self.objct

    def save(self):
        """
        Saves specified object to DB if it doesnt exist yet
        :return: django object saved to db or False if nothing was saved
        """

        if not self.created:
            return False

        exist = genUri.objects.filter(uri=self.uri)
        if exist.count() > 0:
            if exist[0].entity is not None:
                return exist[0].entity
        self.objct.status = 'distinct'
        for obj in self._foreign_keys:
            try:
                attr2, created = obj[1].objects.get_or_create(**obj[2])
                setattr(self.objct, obj[0], attr2)
            except MultipleObjectsReturned:
                pass
        self.objct.save()
        def_coll, created = Collection.objects.get_or_create(name='Default import collection')
        self.objct.collection.add(def_coll)
        self.saved = True
        self._objct_uri.entity = self.objct
        self._objct_uri.save()
        self._add_same_as()
        lab_new = []
        for lab in self.labels:
            lt, created = LabelType.objects.get_or_create(name=lab[2])
            l1 = Label.objects.create(label=lab[0], label_type=lt, temp_entity=self.objct)
            lab_new.append(l1)
        self.labels = lab_new
        rel_obj_new = []
        for obj in self._m2m:
            try:
                attr2, created = obj[1].objects.get_or_create(**obj[2])
                getattr(self.objct, obj[0]).add(attr2)
            except MultipleObjectsReturned:
                pass
        for obj in self.related_objcts:
            for u3 in obj[1]:
                ent1 = RDFParser(u3, obj[2], uri_check=self._uri_check, preserve_uri_minutes=self._preserve_uri_minutes).get_or_create(depth=0)
                if obj[2].lower() == self.kind.lower():
                    mod = ContentType.objects.get(model=f"{self.kind.lower()*2}", app_label=self._app_label_relations).model_class()()
                    setattr(mod, 'related_' + self.kind.lower() + 'A_id', self.objct.pk)
                    setattr(mod, 'related_' + self.kind.lower() + 'B_id', ent1.pk)
                else:
                    mod = ContentType.objects.filter(model__icontains=obj[2], app_label=self._app_label_relations).get(
                        model__icontains=self.kind).model_class()()
                    setattr(mod, 'related_' + self.kind.lower() + '_id', self.objct.pk)
                    setattr(mod, 'related_' + obj[2].lower() + '_id', ent1.pk)
                voc = mod._meta.get_field('relation_type').related_model
                voc1, created = voc.objects.get_or_create(name=obj[0])
                setattr(mod, 'relation_type_id', voc1.pk)
                mod.save()
                rel_obj_new.append(mod)
        self.related_objcts = rel_obj_new
        return self.objct

    @staticmethod
    def _sparql_to_pandas(sparql):
        """
        Converts a RDFlib SPARQL query to Pandas DataFrame
        :param sparql: (object) RDFlib query object
        :return: (DataFrame) Pandas DataFrame of return values
        """
        sp2 = sparql._get_bindings()
        for idx, s2 in enumerate(sp2):
            sp2[idx] = {str(key): str(value) for (key, value) in s2.items()}
        return sp2

    def _prep_string(self, string, regex, linked=False, data_type=False):
        """
        Function that converts SPARQL return value to string used in Django attributes
        :param string: (str) Format string with values ingested from SPARQL query (via Pandas DF)
        :param regex: (tuple) tuple of regex and group to use (regex, group)
        :return: (str) converted string
        """
        conv_mapping = {
            'str': str,
            'string': str,
            'float': float,
            'int': int,
            'integer': int
        }
        if string is None:
            return None
        if isinstance(string, str):
            if regex:
                m = re.search(regex[0], string)
                if m:
                    string = m.group(regex[1])
                else:
                    return None
            if len(string) > 255:
                string = string[:250] + '...'
            if linked:
                g1 = rdflib.Graph()
                g1.parse(string)
                pref1 = (SKOS.prefLabel, RDFS.label)
                pref1 += tuple([URIRef(x) for x in self._settings['matching']['prefLabels']])
                string = g1.preferredLabel(URIRef(string), labelProperties=pref1)
                if len(string) > 0:
                    string = str(string[0][1])
            string = unicodedata.normalize("NFC", string)
        if data_type:
            string = conv_mapping[data_type](string)
        return string

    def _normalize_uri(self, uri):
        """
        Normalizes URIs to canonical form
        :param uri: (url) URI to normalize
        :return: (url) converted URI
        """
        sett = yaml.load(open(self._uri_settings_file, 'r'))
        return clean_uri(sett, uri)

    def merge(self, m_obj, app_label_relations='apis_relations'):
        """
        :param m_obj: the object to merge with (must be an django model object instance)
        :param app_label_relations: (string) the label of the Django app that contains the relations
        :return: django object saved to db or False if nothing was saved
        """
        for rel in ContentType.objects.filter(app_label=app_label_relations, model__icontains=self.kind.lower()):
            rel_q = {'related_' + self.kind.lower(): m_obj}
            rel2 = rel.model_class()
            try:
                for rel_exst in rel2.objects.filter(**rel_q):
                    setattr(rel_exst, 'related_'+self.kind.lower()+'_id', self.objct.pk)
                    rel_exst.save()
            except FieldError:  # e.g. PlacePlace relations have different related_ fields
                rel_q = {'related_' + self.kind.lower()+'A': m_obj}
                for rel_exst in rel2.objects.filter(**rel_q):
                    setattr(rel_exst, 'related_'+self.kind.lower()+'A_id', self.objct.pk)
                    rel_exst.save()
                rel_q = {'related_' + self.kind.lower() + 'B': m_obj}
                for rel_exst in rel2.objects.filter(**rel_q):
                    setattr(rel_exst, 'related_'+self.kind.lower()+'B_id', self.objct.pk)
                    rel_exst.save()
        for z in genUri.objects.filter(entity=m_obj):
            z.entity_id = self.objct.pk
            z.save()
        for z in Label.objects.filter(temp_entity=m_obj):
            z.temp_entity_id = self.objct.pk
            z.save()
        if hasattr(m_obj, 'first_name'):
            legacy_name = '{}, {}'.format(m_obj.name, m_obj.first_name)
        else:
            legacy_name = m_obj.name
        lt, created = LabelType.objects.get_or_create(name='legacy name')
        Label.objects.create(temp_entity_id=self.objct.pk, label=legacy_name, label_type=lt)
        for col in m_obj.collection.all():
            self.objct.collection.add(col)
        if 'apis_highlighter' in settings.INSTALLED_APPS:
            for ann in m_obj.annotation_set.all():  # Todo: check if this works now with highlighter
                ann.entity_link.remove(m_obj)
                ann.entity_link.add(self.objct)
        for txt in m_obj.text.all():
            self.objct.text.add(txt)
        if m_obj.source:
            self.objct.source = m_obj.source
            self.objct.save()
        m_obj.delete()
        return self.objct


    def create_objct(self, depth=2):
        """
        Uses parsed attributes to create an object that is not yet persisted to the db.
        Stores the object in self.objct
        :param depth: (int) depth to follow
        """
        c_dict = dict()
        for s in self._settings['matching']['attributes'].keys():
            if 'domain' in self._settings['matching']['attributes'][s].keys():
                if self._settings['matching']['attributes'][s]['domain'] not in self.uri:
                    print(f"continue: {s} / {self.uri} / {self._settings['matching']['attributes'][s]['domain']}")
                    continue
            access = self._settings['matching']['attributes'][s].get('accessor', None)
            field_name = self._settings['matching']['attributes'][s].get('field name', s)
            fields_1 = self.objct._meta.get_field(field_name)
            local_regex = self._settings['matching']['attributes'][s].get('regex', None)
            local_linked = self._settings['matching']['attributes'][s].get('linked', None)
            local_data_type = self._settings['matching']['attributes'][s].get('data type', False)
            id_1 = self._settings['matching']['attributes'][s].get('identifier', s).split('.')[0]
            if isinstance(fields_1, TCharField) or isinstance(fields_1, TFloatField):
                data = dict()
                if id_1 in self._attributes.keys():
                    df = self._attributes[id_1]
                    cols = [x for x in df.columns if x != 'lang']
                    for c in cols:
                        data[c] = df.at[0, c]
                    print(f"data: {data}")
                    if 'string' in self._settings['matching']['attributes'][s].keys():
                        local_string = fmt.format(self._settings['matching']['attributes'][s]['string'], **data)
                    else:
                        local_string = fmt.format("{"+self._settings['matching']['attributes'][s]['identifier'].split('.')[-1]+"}", **data)
                    c_dict[field_name] = self._prep_string(local_string, local_regex, local_linked, local_data_type)
            elif isinstance(fields_1, TForeignKey):
                data = dict()
                c_dict_f = dict()
                if id_1 in self._attributes.keys():
                    df = self._attributes[id_1]
                    cols = [x for x in df.columns if x != 'lang']
                    for c in cols:
                        data[c] = df.at[0, c]
                    local_string = fmt.format(self._settings['matching']['attributes'][s]['string'], **data)
                    c_dict_f[access] = self._prep_string(local_string, local_regex, local_linked, local_data_type)
                self._foreign_keys.append((field_name, self.objct._meta.get_field(s).related_model, c_dict_f))
            elif isinstance(fields_1, TManyToMany):
                df = self._attributes[id_1]
                for idx, row in df.iterrows():
                    data = dict()
                    if id_1 in self._attributes.keys():
                        cols = [x for x in df.columns if x != 'lang']
                        for c in cols:
                            data[c] = row[c]
                        c_dict_f = dict()
                        local_string = fmt.format(self._settings['matching']['attributes'][s]['string'], **data)
                        c_dict_f[access] = self._prep_string(local_string, local_regex, local_linked, local_data_type)
                    self._m2m.append((field_name, self.objct._meta.get_field(s).related_model, c_dict_f))
        if 'labels' in self._settings['matching'].keys():
            for lab in self._settings['matching']['labels']:
                at1 = lab['identifier'].split('.')
                if at1[0] in self._attributes.keys():
                    if at1[-1] not in self._attributes[at1[0]].columns:
                        print('label_missing')
                        continue
                    if 'lang' not in self._attributes[at1[0]]:
                        self._attributes[at1[0]]['lang'] = 'deu'
                    u2 = self._attributes[at1[0]][[at1[-1], 'lang']]
                    for idx, row in u2.iterrows():
                        self.labels.append((row[at1[-1]], row['lang'], lab['label type']))

        if depth > 0 and 'linked objects' in self._settings['matching'].keys():
            for v in self._settings['matching']['linked objects']:
                at1 = v['identifier'].split('.')
                if at1[0] in self._attributes.keys():
                    if at1[-1] in self._attributes[at1[0]].columns:
                        u2 = self._attributes[at1[0]][at1[-1]].tolist()
                        self.related_objcts.append((v['kind'], u2, v['type']))
        self.objct = self.objct(**c_dict)

    def __init__(self, uri, kind, app_label_entities="apis_entities", app_label_relations="apis_relations",
                 app_label_vocabularies="apis_vocabularies", rdf_settings=APIS_RDF_YAML_SETTINGS,
                 uri_settings=APIS_RDF_URI_SETTINGS, preserve_uri_minutes=5, use_preferred=False, uri_check=True, **kwargs):
        """
        :param uri: (url) Uri to parse the object from (http://test.at). The uri must start with a base url mentioned in the RDF parser settings file.
        :param kind: (string) Kind of entity (Person, Place, Institution, Work, Event)
        :param app_label_entities: (string) Name of the Django app that contains the entities that we create.
        :param app_label_relations: (string) Name of the Django app that contains the relations for the merging process.
        :param app_label_vocabularies: (string) Name of the Django app that contains the vocabularies defining the entities and relations.
        :param use_preferred: (boolean) if True forwards to preferred sources defined in sameAs
        """

        self._clean_uri_store()
        self._rdf_settings_file = rdf_settings
        self._uri_settings_file = uri_settings
        self._preserve_uri_minutes = preserve_uri_minutes
        self._uri_check = uri_check
        self._use_preferred = use_preferred
        self.objct = ContentType.objects.get(app_label=app_label_entities, model__iexact=kind).model_class()
        self._app_label_relations = app_label_relations
        self.kind = kind
        self._foreign_keys = []
        self._m2m = []
        self.uri = self._normalize_uri(uri)
        self._objct_uri = Uri(uri=self.uri)
        self._subject = URIRef(self.uri)
        self.related_objcts = []
        force = kwargs.get('force', None)
        self.labels = []
        self.related_objcts = []

        self.saved = False
        test = self._exist(self.uri, uri_check=uri_check)
        if test[0] and not force:
            self.objct = test[1]
            self.created = False
            print('not created')
        else:
            self.created = True
            o = self._parse()



from django.conf import settings
from django.conf.urls import include, url
from django.contrib import admin
from django.contrib.auth.decorators import login_required
from django.contrib.contenttypes.models import ContentType
from django.urls import path
from django.views.static import serve
from django_filters.rest_framework import DjangoFilterBackend
from rest_framework import response, schemas
from rest_framework import routers
from rest_framework.decorators import api_view, renderer_classes
from rest_framework.schemas import get_schema_view

from apis_core.api_routers import views
from apis_core.apis_entities.api_views import (
    NetJsonViewSet,
    PlaceGeoJsonViewSet,
)
from apis_core.apis_vocabularies.api_views import UserViewSet
from apis_core.helper_functions.ContentType import GetContentTypes

app_name = "apis_core"

router = routers.DefaultRouter()
for app_label, ent in GetContentTypes().get_names():
    try:
        router.register(
            r"{}/{}".format(app_label.split("_")[1], ent.lower()),
            views[ent.lower()],
            ent.lower(),
        )
    except Exception as e:
        print("{} not found, skipping".format(ent.lower()))


if "apis_highlighter" in settings.INSTALLED_APPS:
    from apis_highlighter.api_views import (
        HighlighterProjectViewSet,
        HighlighterTextHighViewSet,
        HighlighterMenuEntryViewSet,
        HighlighterHighlightTextViewSet,
        HighlighterVocabularyAPIViewSet,
        HighlighterAnnotationViewSet,
    )

    router.register(r"HLProjects", HighlighterProjectViewSet)
    router.register(r"HLTextHigh", HighlighterTextHighViewSet)
    router.register(r"HLMenuEntry", HighlighterMenuEntryViewSet)
    # router.register(
    #    r"HLTextHighlighter", HighlighterHighlightTextViewSet, "HLTextHighlighter"
    # )
    # router.register(r"HLVocabularyAPI", HighlighterVocabularyAPIViewSet)
    router.register(r"HLAnnotation", HighlighterAnnotationViewSet)

router.register(r"users", UserViewSet)
router.register(r"GeoJsonPlace", PlaceGeoJsonViewSet, "PlaceGeoJson")
# router.register(r"NetJson", NetJsonViewSet, "NetJson")


from rest_framework import permissions

# from drf_yasg.views import get_schema_view as get_schema_view2
# from drf_yasg import openapi
from drf_spectacular.views import (
    SpectacularAPIView,
    SpectacularSwaggerView,
    SpectacularRedocView,
)


# schema_view2 = get_schema_view2(
#   openapi.Info(
#      title="APIS API",
#      default_version='v1',
#      description="Hyperlinked API of the APIS Framework",
# terms_of_service="https://www.google.com/policies/terms/",
#      contact=openapi.Contact(email="matthias.schloegl@oeaw.ac.at"),
#      license=openapi.License(name="MIT"),
#   ),
#   public=True,
#   permission_classes=(permissions.AllowAny,),
# )

# class APISSchemaGenerator(OpenAPISchemaGenerator):
#    info = "APIS test"
#    title = "APIS_API v2"

#    def __init__(self, *args, **kwargs):
#        super(APISSchemaGenerator, self).__init__(*args, **kwargs)


# class SchemaViewSwagger(schema_view2):
#    generator_class = APISSchemaGenerator

#    def get_filter_parameters(self, filter_backend):
#       if isinstance(filter_backend, DjangoFilterBackend):
#            result = super(SchemaViewSwagger, self).get_filter_parameters(filter_backend)
#            for param in result:
#                if not param.get('description', ''):
#                    param.description = "Filter the returned list by {field_name}".format(field_name=param.name)

#            return result

#        return NotHandled

#    def get_operation(self, operation_keys):
#        super(SchemaViewSwagger, self).get_operation(operation_keys)

#    def __init__(self, *args, **kwargs):
#        super(SchemaViewSwagger, self).__init__(*args, **kwargs)


def build_apis_mock_request(method, path, view, original_request, **kwargs):
    # default mock request
    request = build_mock_request(method, path, view, original_request, **kwargs)
    # the added wagtail magic
    request.router = router
    return request


urlpatterns = [
    url(r"^admin/", admin.site.urls),
    # url(r'^swagger(?P<format>\.json|\.yaml)$', SchemaViewSwagger.without_ui(cache_timeout=-1), name='schema-json'),
    # url(r'^swagger/$', SchemaViewSwagger.with_ui('swagger', cache_timeout=-1), name='schema-swagger-ui'),
    # url(r'^redoc/$', SchemaViewSwagger.with_ui('redoc', cache_timeout=-1), name='schema-redoc'),
    path("swagger/schema/", SpectacularAPIView.as_view(), name="schema"),
    # Optional UI:
    path(
        "swagger/schema/swagger-ui/",
        SpectacularSwaggerView.as_view(url_name="apis_core:schema"),
        name="swagger-ui",
    ),
    path(
        "swagger/schema/redoc/",
        SpectacularRedocView.as_view(url_name="apis_core:schema"),
        name="redoc",
    ),
    url(r"labels/", include("apis_core.apis_labels.urls", namespace="apis_labels")),
    url(r"tei/", include("apis_core.apis_tei.tei_urls", namespace="apis_tei")),
    url(
        r"entities/", include("apis_core.apis_entities.urls", namespace="apis_entities")
    ),
    url(r"visuals/", include("apis_core.apis_vis.urls", namespace="apis_visuals")),
    url(
        r"relations/",
        include("apis_core.apis_relations.urls", namespace="apis_relations"),
    ),
    url(
        r"vocabularies/",
        include("apis_core.apis_vocabularies.urls", namespace="apis_vocabularies"),
    ),
    url(
        r"metainfo/",
        include("apis_core.apis_metainfo.urls", namespace="apis_metainfo"),
    ),
    url(
        r"metainfo-ac/",
        include("apis_core.apis_metainfo.dal_urls", namespace="apis_metainfo-ac"),
    ),
    # url(r'^autocomplete/', include('autocomplete_light.urls')),
    url(
        r"^api/", include((router.urls, "apis_core"), namespace="apis_api")
    ),  # routers do not support namespaces out of the box
    # path('openapi-2', schema_view),
    # path('openapi-api', get_schema_view(
    #    title="APIS",
    #    description="APIS API schema definition",
    #    urlconf='apis_core.apis_entities.api_urls',
    # ), name='openapi-schema-api'),
    url(r"^api2/", include("apis_core.apis_entities.api_urls", namespace="apis_api2")),
    url(r"^api-auth/", include("rest_framework.urls", namespace="rest_framework")),
    # url(r'^api-schema/', schema_view),
    url(r"^apis-vis/", include("apis_core.apis_vis.urls", namespace="apis_vis")),
    url(
        r"^docs/(?P<path>.*)$",
        login_required(serve),
        {"document_root": "apis-core/docs/_build/html"},
        "docs",
    ),
    # url(r'^docs/', include('sphinxdoc.urls')),
    # url(r'^accounts/', include('registration.backends.simple.urls')),
]

if "apis_highlighter" in settings.INSTALLED_APPS:
    urlpatterns.append(
        url(r"highlighter/", include("apis_highlighter.urls", namespace="highlighter"))
    )

if "apis_fulltext_download" in settings.INSTALLED_APPS:

    urlpatterns.append(
        url(
            r"fulltext_download/",
            include("apis_fulltext_download.urls", namespace="apis_fulltext_download"),
        )
    )

if settings.DEBUG:
    if "debug_toolbar" in settings.INSTALLED_APPS:
        import debug_toolbar

        urlpatterns = [url(r"^__debug__/", include(debug_toolbar.urls))] + urlpatterns



from django.conf.urls import url

from . import rel_views
from . import views

app_name = 'apis_relations'

urlpatterns = [
    url(r'^ajax/get/$', views.get_form_ajax, name='get_form_ajax'),
    url(
        r'^ajax/save/(?P<entity_type>\w+)/(?P<kind_form>\w+)/(?P<SiteID>[0-9]+)(?:/(?P<ObjectID>[0-9]*))?/$',
        views.save_ajax_form, name='save_ajax_form'
    ),
    url(r'^(?P<entity>[a-z]+)/list/$',
        rel_views.GenericRelationView.as_view(), name='generic_relations_list'),
    url(r'^(?P<entity>[a-z]+)/(?P<pk>[0-9]+)/detail$',
        rel_views.GenericRelationDetailView.as_view(),
        name='generic_relations_detail_view'),
]



import inspect
import re
import sys
import unicodedata
import yaml

# from reversion import revisions as reversion
import reversion
from django.conf import settings
from django.contrib.auth.models import Group
from django.db import models
from django.db.models import Q
from django.db.models.signals import m2m_changed, post_save
from django.dispatch import receiver
from django.urls import reverse
from guardian.shortcuts import assign_perm, remove_perm

from apis_core.apis_metainfo.models import Collection, TempEntityClass, Uri
from apis_core.apis_vocabularies.models import (
    EventType,
    InstitutionType,
    PlaceType,
    ProfessionType,
    Title,
    WorkType,
)
from apis_core.helper_functions import EntityRelationFieldGenerator

BASE_URI = getattr(settings, "APIS_BASE_URI", "http://apis.info/")


class AbstractEntity(TempEntityClass):
    """
    Abstract super class which encapsulates common logic between the different entity kinds and provides various methods
    relating to either all or a specific entity kind.

    Most of the class methods are designed to be used in the sublcass as they are considering contexts which depend on
    the subclass entity type. So they are to be understood in that dynamic context.
    """

    # Placeholder for list filter classes attached to each entity later
    list_filter_class = None

    class Meta:
        abstract = True

    def __init__(self, *args, **kwargs):

        super().__init__(*args, **kwargs)
        self.__class__.create_relation_methods_from_manytomany_fields()

    # Methods dealing with individual data retrievals of instances
    ####################################################################################################################

    def __str__(self):

        if self.__class__ == Person:

            def valid(name):
                return name != "" and name is not None

            if valid(self.first_name) and valid(self.name):
                return "{}, {}".format(self.name, self.first_name)
            elif valid(self.first_name) and not valid(self.name):
                return "{}, {}".format("no surename provided", self.first_name)
            elif not valid(self.first_name) and valid(self.name):
                return self.name
            elif not valid(self.first_name) and not valid(self.name):
                return "no name provided"

        else:

            if self.name != "":
                return self.name
            else:
                return "no name provided"

    @classmethod
    def get_or_create_uri(cls, uri):
        uri = str(uri)
        try:
            if re.match(r"^[0-9]*$", uri):
                p = cls.objects.get(pk=uri)
            else:
                p = cls.objects.get(uri__uri=uri)
            return p
        except:
            print("Found no object corresponding to given uri.")
            return False

    @classmethod
    def set_list_filter_class(cls, list_filter_class):
        cls.list_filter_class = list_filter_class

    # Various Methods enabling convenient shortcuts between entities, relations, fields, etc
    ####################################################################################################################

    @classmethod
    def create_relation_methods_from_manytomany_fields(cls):
        """
        Creates methods on an entity class with which other related entities can be called without the need to consider
        potential self-references (the A and B sides therein).

        The resulting methods follow the syntax:
        <entity>.get_related_<entity>_instances()

        e.g. for Person:
        person.get_related_work_instances()
        or
        person.get_related_person_instances()

        Note that with these methods it is not necessary to differentiate between A and B entities when self-relations exist.

        The result of any such method call is the queryset of the related entities.
        (And not a ManyToManyManager as is the case when calling <entity>.<entity>_set where in the case of self-relation
        it must be differentiated between A and B entities, e.g. person.personA_set )

        It was not possible to my understanding to change managers in such a way that two (the A and the B) could be combined
        into one manager. Hence these additional shortcut methods.

        :return: None
        """

        def create_function_from_manytomany_field_to_other_entity(
            entity_manytomany_field,
        ):
            """
            creates the individual method from a ManyToMany field by calling the manager's objects.all()

            This method creation has to be done here in a separate method, so that it can be called once before assignment
            as otherwise the variable 'entity_name' in the loop below changes with each iteration and with that also the
            method references (due to python's "late binding").
            A method call in between thus forces the content of 'entity_name' to be assigned for good to the
            respective class ( = forced early binding).
            For more details on this: https://stackoverflow.com/questions/3431676/creating-functions-in-a-loop

            :param entity_manytomany_field: the ManyToManyManager to another model
            :return: method which will call the managers's objects.all() method
            """
            return lambda self: getattr(self, entity_manytomany_field).all().distinct()

        def create_function_from_manytomany_field_to_self_entity(
            entityA_manytomany_field, entityB_manytomany_field
        ):
            """
            Same method as above, but with two managers instead of one for the case of self-relations.

            Both managers' objects.all() methods are called and their queryset results are unionised for the
            shortcut method of an entity to its own related entities.

            :param entityA_manytomany_field: ManyToManyManager to entity A in a self-relation
            :param entityB_manytomany_field: ManyToManyManager to entity A in a self-relation
            :return: method to call both and return the distinct union of them
            """
            return lambda self: (
                getattr(self, entityA_manytomany_field)
                .all()
                .union(getattr(self, entityB_manytomany_field).all())
            ).distinct()

        for entity_name in cls.get_all_entity_names():
            # Iterate over each entity defined within this models' module

            related_entity_function_name = "get_related_" + entity_name + "_instances"

            if not hasattr(cls, related_entity_function_name):

                if cls.__name__.lower() == entity_name:
                    # If the related entity is the same as this current one, then set the names of the related functions
                    # to A and B and also combine them into one function where both A and B are returned.

                    related_entityA_function_name = (
                        "get_related_" + entity_name + "A_instances"
                    )
                    related_entityB_function_name = (
                        "get_related_" + entity_name + "B_instances"
                    )
                    entityA_manytomany_field = entity_name + "A_set"
                    entityB_manytomany_field = entity_name + "B_set"

                    setattr(
                        cls,
                        related_entityA_function_name,
                        create_function_from_manytomany_field_to_other_entity(
                            entityA_manytomany_field
                        ),
                    )

                    setattr(
                        cls,
                        related_entityB_function_name,
                        create_function_from_manytomany_field_to_other_entity(
                            entityB_manytomany_field
                        ),
                    )

                    setattr(
                        cls,
                        related_entity_function_name,
                        create_function_from_manytomany_field_to_self_entity(
                            entityA_manytomany_field, entityB_manytomany_field
                        ),
                    )

                else:
                    # If the related entity is a different one, then just build on the usual names

                    entity_manytomany_field = entity_name + "_set"

                    setattr(
                        cls,
                        related_entity_function_name,
                        create_function_from_manytomany_field_to_other_entity(
                            entity_manytomany_field
                        ),
                    )

    # Methods dealing with all entities
    ####################################################################################################################

    _all_entity_classes = None
    _all_entity_names = None

    @classmethod
    def get_all_entity_classes(cls):
        """
        :return: list of all python classes of the entities defined within this models' module
        """

        if cls._all_entity_classes == None:

            entity_classes = []
            entity_names = []

            for entity_name, entity_class in inspect.getmembers(
                sys.modules[__name__], inspect.isclass
            ):

                if (
                    entity_class.__module__ == "apis_core.apis_entities.models"
                    and entity_name != "ent_class"
                    and entity_name != "AbstractEntity"
                ):

                    entity_classes.append(entity_class)
                    entity_names.append(entity_name.lower())

            cls._all_entity_classes = entity_classes
            cls._all_entity_names = entity_names

        return cls._all_entity_classes

    @classmethod
    def get_entity_class_of_name(cls, entity_name):
        """
        :param entity_name: str : The name of an entity
        :return: The model class of the entity respective to the given name
        """

        for entity_class in cls.get_all_entity_classes():
            if entity_class.__name__.lower() == entity_name.lower():
                return entity_class

        raise Exception("Could not find entity class of name:", entity_name)

    @classmethod
    def get_all_entity_names(cls):
        """
        :return: list of all class names in lower case of the entities defined within this models' module
        """

        if cls._all_entity_names == None:

            cls.get_all_entity_classes()

        return cls._all_entity_names

    # Methods dealing with related entities
    ####################################################################################################################

    _related_entity_field_names = None

    @classmethod
    def get_related_entity_field_names(cls):
        """
        :return: a list of names of all ManyToMany field names relating to entities from the respective entity class

        E.g. for Person.get_related_entity_field_names() or person_instance.get_related_entity_field_names() ->
        ['event_set', 'institution_set', 'personB_set', 'personA_set', 'place_set', 'work_set']

        Note: this method depends on the 'generate_all_fields' function of the EntityRelationFieldGenerator class 
        which wires the ManyToMany Fields into the entities and respective relationtypes. 
        This method is nevertheless defined here within AbstractEntity for documentational purpose.
        """

        if cls._related_entity_field_names == None:
            raise Exception("_related_entity_field_names was not initialized yet.")
        else:
            return cls._related_entity_field_names

    @classmethod
    def add_related_entity_field_name(cls, entity_field_name):
        """
        :param entity_field_name: the name of one of several ManyToMany fields created automatically
        :return: None

        Note: this method depends on the 'generate_all_fields' function of the EntityRelationFieldGenerator class 
        which wires the ManyToMany Fields into the entities and respective relationtypes. 
        This method is nevertheless defined here within AbstractEntity for documentational purpose.
        """

        if cls._related_entity_field_names == None:
            cls._related_entity_field_names = []

        cls._related_entity_field_names.append(entity_field_name)

    def get_related_entity_instances(self):
        """
        :return: list of queryset of all entity instances which are somehow related to the calling entity instance
        """

        queryset_list = []

        for entity_name in self.get_all_entity_names():

            queryset = getattr(self, "get_related_" + entity_name + "_instances")()
            if len(queryset) > 0:
                queryset_list.append(queryset)

        return queryset_list

    # Methods dealing with related relations
    ####################################################################################################################

    @classmethod
    def get_related_relation_classes(cls):
        """
        :return: list of python classes of the relations which are related to the respective entity class

        E.g. for Place.get_related_relation_classes() or place_instance.get_related_relation_classes() ->
        [ InstitutionPlace, PersonPlace, PlaceEvent, PlacePlace, PlaceWork ]
        """

        # TODO __sresch__ : check for best practice on local imports vs circularity problems.
        from apis_core.apis_relations.models import AbstractRelation

        return AbstractRelation.get_relation_classes_of_entity_class(cls)

    @classmethod
    def get_related_relation_field_names(cls):
        """
        :return: list of class names in lower case of the relations which are related to the respective entity class

        E.g. for Place.get_related_relation_names() or place_instance.get_related_relation_names() ->
        ['institutionplace_set', 'personplace_set', 'placeevent_set', 'placeplace_set', 'placework_set']
        """
        # TODO __sresch__ : check for best practice on local imports vs circularity problems.
        from apis_core.apis_relations.models import AbstractRelation

        return AbstractRelation.get_relation_field_names_of_entity_class(cls)

    def get_related_relation_instances(self):
        """
        :return: list of queryset of all relation instances which are somehow related to the calling entity instance
        """

        queryset_list = []

        for relation_class in self.get_related_relation_classes():

            q_args = Q()

            if relation_class.get_related_entity_classA() == self.__class__:
                q_args |= Q(**{relation_class.get_related_entity_field_nameA(): self})

            if relation_class.get_related_entity_classB() == self.__class__:
                q_args |= Q(**{relation_class.get_related_entity_field_nameB(): self})

            queryset = relation_class.objects.filter(q_args)
            queryset_list.extend(list(queryset))

        return queryset_list

    # Methods dealing with related relationtypes
    ####################################################################################################################

    _related_relationtype_classes = None
    _related_relationtype_field_names = None
    _related_relationtype_names = None

    @classmethod
    def get_related_relationtype_classes(cls):
        """
        :return: list of python classes of the relation types which are related to the respective entity class

        E.g. for Place.get_related_relation_classes() or place_instance.get_related_relation_classes() ->
        [ InstitutionPlaceRelation, PersonPlaceRelation, PlaceEventRelation, PlacePlaceRelation, PlaceWorkRelation ]
        """

        if cls._related_relationtype_classes == None:

            relationtype_classes = []
            relationtype_names = []

            # TODO __sresch__ : check for best practice on local imports vs circularity problems.
            from apis_core.apis_vocabularies.models import AbstractRelationType

            for (
                relationtype_class
            ) in AbstractRelationType.get_all_relationtype_classes():

                relationtype_name = relationtype_class.__name__.lower()

                if cls.__name__.lower() in relationtype_name:

                    relationtype_classes.append(relationtype_class)
                    relationtype_names.append(relationtype_name)

            cls._related_relationtype_classes = relationtype_classes
            cls._related_relationtype_names = relationtype_names

        return cls._related_relationtype_classes

    @classmethod
    def get_related_relationtype_names(cls):
        """
        :return: list of class names in lower case of the relation types which are related to the respective entity class

        E.g. for Place.get_related_relation_classes() or place_instance.get_related_relation_classes() ->
        [ 'institutionplacerelation', 'personplacerelation', 'placeeventrelation', 'placeplacerelation', 'placeworkrelation' ]
        """

        if cls._related_relationtype_names == None:

            cls.get_related_relationtype_classes()

        return cls._related_relationtype_names

    @classmethod
    def get_related_relationtype_field_names(cls):
        """
        :return: a list of names of all ManyToMany field names relating to relationtypes from the respective entity class

        E.g. for PersonPerson.get_related_relationtype_field_names() or person_instance.get_related_relationtype_field_names() ->
        ['event_relationtype_set', 'institution_relationtype_set', 'personB_relationtype_set', 'personA_relationtype_set', 'place_relationtype_set', 'work_relationtype_set']

        Note: this method depends on the 'generate_all_fields' function of the EntityRelationFieldGenerator class 
        which wires the ManyToMany Fields into the entities and respective relationtypes. 
        This method is nevertheless defined here within AbstractEntity for documentational purpose.
        """

        if cls._related_relationtype_field_names == None:
            raise Exception(
                "_related_relationtype_field_names was not initialized yet."
            )
        else:
            return cls._related_relationtype_field_names

    @classmethod
    def add_related_relationtype_field_name(cls, relationtype_field_name):
        """
        :param entity_field_name: the name of one of several ManyToMany fields created automatically
        :return: None

        Note: this method depends on the 'generate_all_fields' function of the EntityRelationFieldGenerator class 
        which wires the ManyToMany Fields into the entities and respective relationtypes. 
        This method is nevertheless defined here within AbstractEntity for documentational purpose.
        """

        if cls._related_relationtype_field_names == None:
            cls._related_relationtype_field_names = []

        cls._related_relationtype_field_names.append(relationtype_field_name)

    def get_related_relationtype_instances(self):
        """
        :return: list of queryset of all relationtype instances which are somehow related to the calling entity instance
        """

        queryset_list = []

        for entity_name in self.get_all_entity_names():
            queryset = None

            if entity_name != self.__class__.__name__.lower():

                queryset = (
                    getattr(self, entity_name + "_relationtype_set").all().distinct()
                )

            else:

                querysetA = (
                    getattr(self, entity_name + "A_relationtype_set").all().distinct()
                )
                querysetB = (
                    getattr(self, entity_name + "B_relationtype_set").all().distinct()
                )
                queryset = querysetA.union(querysetB)

            if queryset and len(queryset) > 0:
                queryset_list.append(queryset)

        return queryset_list


@reversion.register(follow=["tempentityclass_ptr"])
class Person(AbstractEntity):

    GENDER_CHOICES = (
        ("female", "female"),
        ("male", "male"),
        ("third gender", "third gender"),
    )
    first_name = models.CharField(
        max_length=255,
        help_text="The persons´s forename. In case of more then one name...",
        blank=True,
        null=True,
    )
    profession = models.ManyToManyField(ProfessionType, blank=True)
    title = models.ManyToManyField(Title, blank=True)
    gender = models.CharField(
        max_length=15, choices=GENDER_CHOICES, blank=True, null=True
    )

    def save(self, *args, **kwargs):
        if self.first_name:
            # secure correct unicode encoding
            if self.first_name != unicodedata.normalize("NFC", self.first_name):
                self.first_name = unicodedata.normalize("NFC", self.first_name)
        super(Person, self).save(*args, **kwargs)
        return self


@reversion.register(follow=["tempentityclass_ptr"])
class Place(AbstractEntity):

    kind = models.ForeignKey(
        PlaceType, blank=True, null=True, on_delete=models.SET_NULL
    )
    lat = models.FloatField(blank=True, null=True, verbose_name="latitude")
    lng = models.FloatField(blank=True, null=True, verbose_name="longitude")

    def save(self, *args, **kwargs):
        if isinstance(self.lat, float) and isinstance(self.lng, float):
            self.status = "distinct"
        super(Place, self).save(*args, **kwargs)
        return self


@reversion.register(follow=["tempentityclass_ptr"])
class Institution(AbstractEntity):

    kind = models.ForeignKey(
        InstitutionType, blank=True, null=True, on_delete=models.SET_NULL
    )


@reversion.register(follow=["tempentityclass_ptr"])
class Event(AbstractEntity):

    kind = models.ForeignKey(
        EventType, blank=True, null=True, on_delete=models.SET_NULL
    )


@reversion.register(follow=["tempentityclass_ptr"])
class Work(AbstractEntity):

    kind = models.ForeignKey(WorkType, blank=True, null=True, on_delete=models.SET_NULL)


a_ents = getattr(settings, "APIS_ADDITIONAL_ENTITIES", False)


def prepare_fields_dict(fields_list, vocabs, vocabs_m2m):
    res = dict()
    for f in fields_list:
        res[f["name"]] = getattr(models, f["field_type"])(**f["attributes"])
    for v in vocabs:
        res[v] = models.ForeignKey(
            f"apis_vocabularies.{v}", blank=True, null=True, on_delete=models.SET_NULL
        )
    for v2 in vocabs_m2m:
        res[v2] = models.ManyToManyField(f"apis_vocabularies.{v2}", blank=True)
    return res


ents_cls_list = []
if a_ents:
    with open(a_ents, "r") as ents_file:
        ents = yaml.load(ents_file, Loader=yaml.CLoader)
        print(ents)
        for ent in ents["entities"]:
            attributes = prepare_fields_dict(
                ent["fields"], ent.get("vocabs", []), ent.get("vocabs_m2m", [])
            )
            attributes["__module__"] = __name__
            ent_class = type(ent["name"], (AbstractEntity,), attributes)
            globals()[ent["name"]] = ent_class
            ents_cls_list.append(ent_class)
            reversion.register(ent_class, follow=["tempentityclass_ptr"])


@receiver(post_save, dispatch_uid="create_default_uri")
def create_default_uri(sender, instance, **kwargs):
    if kwargs["created"] and sender in [Person, Institution, Place, Work, Event] + ents_cls_list:
        if BASE_URI.endswith("/"):
            base1 = BASE_URI[:-1]
        else:
            base1 = BASE_URI
        uri_c = "{}{}".format(
            base1, reverse("GetEntityGenericRoot", kwargs={"pk": instance.pk}),
        )
        uri2 = Uri(uri=uri_c, domain="apis default", entity=instance)
        uri2.save()


lst_entities_complete = [
    globals()[x]
    for x in globals()
    if isinstance(globals()[x], models.base.ModelBase)
    and globals()[x].__module__ == "apis_core.apis_entities.models"
    and x != "AbstractEntity"
    and globals()[x]
]
lst_entities_complete = list(dict.fromkeys(lst_entities_complete))
perm_change_senders = [
    getattr(getattr(x, "collection"), "through") for x in lst_entities_complete
] # TODO: Inspect. This list here will contain only duplicates if `lst_entities_complete` contains all entities


@receiver(
    m2m_changed, dispatch_uid="create_object_permissions",
)
def create_object_permissions(sender, instance, **kwargs):
    if kwargs["action"] == "pre_add" and sender in perm_change_senders:
        perms = []
        for j in kwargs["model"].objects.filter(pk__in=kwargs["pk_set"]):
            perms.extend(j.groups_allowed.all())
        for x in perms:
            assign_perm("change_" + instance.__class__.__name__.lower(), x, instance)
            assign_perm("delete_" + instance.__class__.__name__.lower(), x, instance)
    elif kwargs["action"] == "post_remove" and sender in perm_change_senders:
        perms = []
        perms_keep = []
        for j in kwargs["model"].objects.filter(pk__in=kwargs["pk_set"]):
            perms.extend(j.groups_allowed.all())
        for u in instance.collection.all():
            perms_keep.extend(u.groups_allowed.all())
        rm_perms = set(perms) - set(perms_keep)
        for x in rm_perms:
            remove_perm("change_" + instance.__class__.__name__.lower(), x, instance)
            remove_perm("delete_" + instance.__class__.__name__.lower(), x, instance)


@receiver(
    m2m_changed,
    sender=Collection.groups_allowed.through,
    dispatch_uid="add_usergroup_collection",
)
def add_usergroup_collection(sender, instance, **kwargs):
    if kwargs["action"] == "pre_add":
        for x in kwargs["model"].objects.filter(pk__in=kwargs["pk_set"]):
            for z in ["change", "delete"]:
                for y in [Person, Institution, Place, Event, Work]:
                    assign_perm(
                        z + "_" + y.__name__.lower(),
                        x,
                        y.objects.filter(collection=instance),
                    )


if "registration" in getattr(settings, "INSTALLED_APPS", []):
    from registration.backends.simple.views import RegistrationView
    from registration.signals import user_registered

    @receiver(
        user_registered,
        sender=RegistrationView,
        dispatch_uid="add_registered_user_to_group",
    )
    def add_user_to_group(sender, user, request, **kwargs):
        user_group = getattr(settings, "APIS_AUTO_USERGROUP", None)
        if user_group is not None:
            user.groups.add(Group.objects.get(name=user_group))


# Call the field generation function here, after all relevant entity classes have been defined above
EntityRelationFieldGenerator.generate_all_fields()



#!/usr/bin/python
# -*- coding: utf-8 -*-
import json
import operator
import re
from functools import reduce

import dateutil.parser
import requests
from dal import autocomplete
from django import http
from django.conf import settings
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import FieldError
from django.db.models import Q

from apis_core.apis_metainfo.models import Uri, Collection
from apis_core.apis_vocabularies.models import VocabsBaseClass
from apis_core.default_settings.NER_settings import autocomp_settings as ac_settings
from .models import AbstractEntity


class CustomEntityAutocompletes(object):
    """A class for collecting all the custom autocomplete functions for one entity.

    Attributes:

    - self.entity: (string) entity types
    - self.more: (boolean) if more results can be fetched (pagination)
    - self.page_size: (integer) page size
    - self.results: (list) results
    - self.query: (string) query string

    Methods:
    - self.more(): fetch more results
    """

    def __init__(self, entity, query, page_size=20, offset=0, *args, **kwargs):
        """
        :param entity: (string) entity type to fetch additional autocompletes for
        """
        func_list = {}
        if entity not in func_list.keys():
            self.results = None
            return None
        res = []
        more = dict()
        more_gen = False
        for x in func_list[entity]:
            res2 = x().query(query, page_size, offset)
            if len(res2) == page_size:
                more[x.__name__] = (True, offset+1)
                more_gen = True
            res.extend(res2)
        self.results = res
        self.page_size = page_size
        self.more = more_gen
        self._more_dict = more
        self.query = query
        self.offset = offset

    def get_more(self):
        """
        Function to retrieve more results.
        """
        res4 = []
        for key, value in self._more_dict.items():
            if value[0]:
                res3 = globals()[key](self.query, self.page_size, value[1])
                if len(res3) == self.page_size:
                    self._more_dict[key] = (True, value[1]+1)
                else:
                    self._more_dict[key] = (False, value[1])
                self.results.extend(res3)
                res4.extend(res3)
        return res4


class GenericEntitiesAutocomplete(autocomplete.Select2ListView):

    @staticmethod
    def parse_stanbol_object(obj, key, *args):
        if len(args) > 0:
            lst1 = args[0]
        else:
            lst1 = None
        if obj[1] == 'GNDDate':
            if lst1 is not None:
                try:
                    return dateutil.parser.parse(lst1[key][0]['value'])
                except:
                    return lst1[key][0]['value']
            else:
                return obj[0]
        elif obj[1] == 'String':
            if lst1 is not None:
                return lst1[key][0]['value']
            else:
                return obj[0]
        elif obj[1] == 'gndLong':
            if lst1 is not None:
                try:
                    return re.search('Point \( [+-]([0-9\.]+) [+-]([0-9\.]+)', lst1[key][0]['value']).group(1)
                except:
                    print('extract fails')
                    return None
            else:
                print('no match')

    def get(self, request, *args, **kwargs):
        page_size = 20
        offset = (int(self.request.GET.get('page', 1))-1)*page_size
        ac_type = self.kwargs['entity']
        db_include = self.kwargs.get('db_include', False)
        choices = []
        headers = {'Content-Type': 'application/json'}
        ent_model = AbstractEntity.get_entity_class_of_name(ac_type)
        if self.q.startswith('http'):
            res = ent_model.objects.filter(uri__uri=self.q.strip())
        elif len(self.q) > 0:
            q1 = re.match('^([^\[]+)\[([^\]]+)\]$', self.q)
            if q1:
                q = q1.group(1).strip()
                q3 = q1.group(2).split(',')
                q3 = [e.strip() for e in q3]
            else:
                q = re.match('^[^\[]+', self.q).group(0)
                q3 = False
            if re.match('^[^*]+\*$', q.strip()):
                search_type = '__istartswith'
                q = re.match('^([^*]+)\*$', q.strip()).group(1)
            elif re.match('^\*[^*]+$', q.strip()):
                search_type = '__iendswith'
                q = re.match('^\*([^*]+)$', q.strip()).group(1)
            elif re.match('^\"[^"]+\"$', q.strip()):
                search_type = ''
                q = re.match('^\"([^"]+)\"$', q.strip()).group(1)
            elif re.match('^[^*]+$', q.strip()):
                search_type = '__icontains'
                q = q.strip()
            else:
                search_type = '__icontains'
                q = q.strip()
            arg_list = [Q(**{x+search_type: q}) for x in settings.APIS_ENTITIES[ac_type.title()]['search']]
            res = ent_model.objects.filter(reduce(operator.or_, arg_list)).distinct()
            if q3:
                f_dict2 = {}
                for fd in q3:
                    f_dict2[fd.split('=')[0].strip()] = fd.split('=')[1].strip()
                try:
                    res = res.filter(**f_dict2)
                except Exception as e:
                    choices.append({'name': str(e)})
        else:
            q = ''
            res = []
        test_db = True
        test_stanbol = False
        test_stanbol_list = dict()
        more = True
        if not db_include:
            for r in res[offset:offset+page_size]:
                f = dict()
                dataclass = ""
                try:
                    f['id'] = Uri.objects.filter(entity=r)[0].uri
                except:
                    continue
                if hasattr(r, 'lng'):
                    if r.lng and r.lat:
                        dataclass = 'data-vis-tooltip="{}" data-lat="{}" \
                        data-long="{}"  class="apis-autocomplete-span"'.format(ac_type, r.lat, r.lng)
                f['text'] = '<span {}><small>db</small> {}</span>'.format(dataclass, str(r))
                choices.append(f)
            if len(choices) < page_size:
                test_db = False
        else:
            test_db = False
        if ac_type.title() in ac_settings.keys():
            for y in ac_settings[ac_type.title()]:
                ldpath = ""
                for d in y['fields'].keys():
                    ldpath += "{} = <{}>;\n".format(d, y['fields'][d][0])
                if self.q.startswith('http'):
                    q = False
                    match_url_geo = re.search(r'geonames[^0-9]+([0-9]+)', self.q.strip())
                    if match_url_geo:
                        url = 'http://sws.geonames.org/{}/'.format(match_url_geo.group(1))
                    else:
                        url = self.q.strip()
                    params = {'id': url, 'ldpath': ldpath}
                    headers = {'Content-Type': 'application/json'}
                    w = requests.get(y['url'].replace('find', 'entity'), params=params, headers=headers)
                    res3 = dict()
                    ldpath_fields = [y['fields'][d][0] for d in y['fields'].keys()]
                    print(w.status_code)
                    if w.status_code == 200:
                        for x in w.json()['representation'].keys():
                            if x in ldpath_fields:
                                for d in y['fields'].keys():
                                    if y['fields'][d][0] == x:
                                        res3[d] = w.json()['representation'][x]
                        res = dict()
                        res3['id'] = w.json()['id']
                        res['results'] = [res3]
                    else:
                        continue
                else:

                    data = {
                        'limit': page_size,
                        'ldpath': ldpath,
                        'offset': offset,
                        'constraints': [{
                            "type": "text",
                            "patternType": "wildcard",
                            "field": "http://www.w3.org/2000/01/rdf-schema#label",
                            "text": q.split()
                        }]
                    }
                    if q3 and 'search fields' in y.keys():
                        for fd in q3:
                            fd = [fd2.strip() for fd2 in fd.split('=')]
                            if fd[0] in y['search fields'].keys():
                                fd3 = y['search fields'][fd[0]]
                                v = False
                                if isinstance(fd3[1], dict):
                                    if fd[1] in fd3[1].keys():
                                        v = fd3[1][fd[1]]
                                else:
                                    v = fd3[1](fd[1])
                                if fd3[2] == 'reference' and v:
                                    fd_4 = {
                                        'type': 'reference',
                                        'value': v,
                                        'field': fd3[0]
                                    }
                                    data['constraints'].append(fd_4)
                                elif fd3[2] == 'date_exact' and v:
                                    fd_4 = {
                                        'type': 'value',
                                        'value': v,
                                        'field': fd3[0],
                                        "datatype": "xsd:dateTime"
                                    }
                                    data['constraints'].append(fd_4)
                                elif fd3[2] == 'date_gt' and v:
                                    fd_4 = {
                                        'type': 'range',
                                        'lowerBound': v,
                                        'upperBound': "2100-12-31T23:59:59.999Z",
                                        'field': fd3[0],
                                        "datatype": "xsd:dateTime"
                                    }
                                    data['constraints'].append(fd_4)
                                elif fd3[2] == 'date_lt' and v:
                                    fd_4 = {
                                        'type': 'range',
                                        'lowerBound': "1-01-01T23:59:59.999Z",
                                        'upperBound': v,
                                        'field': fd3[0],
                                        "datatype": "xsd:dateTime"
                                    }
                                    data['constraints'].append(fd_4)
                            else:
                                choices.append({'name': 'No additional query setting for Stanbol'})
                    try:
                        url2 = y['url'].replace('find', 'query')
                        r = requests.post(url2, data=json.dumps(data), headers=headers)
                        if r.status_code != 200:
                            choices.append({'name': 'Connection to Stanbol failed'})
                            continue
                        res = r.json()
                    except Exception as e:
                        choices.append({'name': 'Connection to Stanbol failed'})
                        print(e)
                        continue
                if len(res['results']) < page_size:
                    test_stanbol_list[y['url']] = False
                else:
                    test_stanbol_list[y['url']] = True
                for x in res['results']:
                    f = dict()
                    dataclass = ""
                    name = x['name'][0]['value']
                    if ac_settings['score'] in x.keys():
                        score = str(x[ac_settings['score']][0]['value'])
                    else:
                        score = 'NA'
                    id = x[ac_settings['uri']]
                    score = score
                    f['id'] = id
                    source = y['source']
                    if 'lat' in x.keys() and 'long' in x.keys():
                        dataclass = 'data-vis-tooltip="{}" \
                        data-lat="{}" data-long="{}"'.format(
                            ac_type, x['lat'][0]['value'], x['long'][0]['value'])
                    if 'descr' in x.keys():
                        descr = x['descr'][0]['value']
                    else:
                        descr = None
                    f['text'] = '<span {} class="apis-autocomplete-span"><small>{}</small> <b>{}</b>\
                    ({}): {}</span>'.format(dataclass, source, name, score, descr)
                    choices.append(f)
            for k in test_stanbol_list.keys():
                if test_stanbol_list[k]:
                    test_stanbol = True
        else:
            test_stanbol = False
        cust_auto_more = False
        if q:
            cust_auto = CustomEntityAutocompletes(ac_type, q, page_size=page_size, offset=offset)
            if cust_auto.results is not None:
                cust_auto_more = cust_auto.more
                if len(cust_auto.results) > 0:
                    choices.extend(cust_auto.results)
        if not test_db and not test_stanbol and not cust_auto_more:
            more = False
        return http.HttpResponse(json.dumps({
            'results': choices + [],
            'pagination': {'more': more}
        }), content_type='application/json')


class GenericVocabulariesAutocomplete(autocomplete.Select2ListView):
    def get(self, request, *args, **kwargs):
        page_size = 20
        offset = (int(self.request.GET.get('page', 1))-1)*page_size
        more = False
        vocab = self.kwargs['vocab']
        direct = self.kwargs['direct']
        q = self.q
        vocab_model = ContentType.objects.get(app_label='apis_vocabularies', model=vocab).model_class()
        if direct == 'normal':
            if vocab_model.__bases__[0] == VocabsBaseClass:
                choices = [{'id': x.pk, 'text': x.label} for x in vocab_model.objects.filter(name__icontains=q).order_by('parent_class__name', 'name')[offset:offset+page_size]]
            else:
                choices = [{'id': x.pk, 'text': x.label} for x in vocab_model.objects.filter(
                    Q(name__icontains=q) | Q(name_reverse__icontains=q)).order_by('parent_class__name', 'name')[offset:offset+page_size]]
        elif direct == 'reverse':
            choices = [{'id': x.pk, 'text': x.label_reverse} for x in vocab_model.objects.filter(
                Q(name__icontains=q) | Q(name_reverse__icontains=q)).order_by('parent_class__name', 'name')[offset:offset+page_size]]
        if len(choices) == page_size:
            more = True
        return http.HttpResponse(json.dumps({
            'results': choices + [],
            'pagination': {'more': more}
        }), content_type='application/json')


class GenericNetworkEntitiesAutocomplete(autocomplete.Select2ListView):
    def get(self, request, *args, **kwargs):
        entity = self.kwargs['entity']
        q = self.q
        if q.startswith('cl:'):
            res = Collection.objects.filter(name__icontains=q[3:])
            results = [{'id': 'cl:'+str(x.pk), 'text': x.name} for x in res]
        elif q.startswith('reg:'):
            results = []
            if entity.lower() == 'person':
                filen = 'reg_persons.json'
            elif entity.lower() == 'place':
                filen = 'reg_places.json'
            with open(filen, 'r') as reg:
                r1 = json.load(reg)
                r_dict = dict()
                for r2 in r1:
                    if q[4:].lower() in r2[1].lower():
                        if r2[1] in r_dict.keys():
                            r_dict[r2[1]] += "|{}".format(r2[0])
                        else:
                            r_dict[r2[1]] = r2[0]
            for k in r_dict.keys():
                results.append({'id': 'reg:'+r_dict[k], 'text': k})

        else:
            ent_model = ContentType.objects.get(
                app_label__startswith='apis_', model=entity
            ).model_class()
            try:
                arg_list = [
                    Q(
                        **{x + '__icontains': q}
                    ) for x in settings.APIS_ENTITIES[entity.title()]['search']
                ]
            except KeyError:
                arg_list = [
                    Q(
                        **{x + '__icontains': q}
                    ) for x in ['name']
                ]
            try:
                res = ent_model.objects.filter(reduce(operator.or_, arg_list)).distinct()
            except FieldError:
                arg_list = [
                    Q(
                        **{x + '__icontains': q}
                    ) for x in ['text']
                ]
                res = ent_model.objects.filter(reduce(operator.or_, arg_list)).distinct()
            results = [{'id': x.pk, 'text': str(x)} for x in res]
        return http.HttpResponse(json.dumps({
            'results': results
        }), content_type='application/json')



gno = 'http://www.geonames.org/ontology#'
wgs84 = "http://www.w3.org/2003/01/geo/wgs84_pos#"
gndo = 'http://d-nb.info/standards/elementset/gnd#'
owl = "http://www.w3.org/2002/07/owl#"
geo = "http://www.opengis.net/ont/geosparql#"

sameAs = [
    "http://schema.org/sameAs",
    "http://www.w3.org/2002/07/owl#sameAs"
]

sett_RDF_generic = {
    'Place': {
        'data': [
            {
                'base_url': 'http://sws.geonames.org/',
                'attributes': [
                    {
                        'name': 'name',
                        'sparql': """
                                PREFIX gn: <http://www.geonames.org/ontology#>
                                SELECT ?prefName (LANG(?prefName) AS ?lang)
                                WHERE {{
                                    <{subject}> gn:officialName ?prefName
                                    FILTER (LANGMATCHES(LANG(?prefName), "de") || LANGMATCHES(LANG(?prefName), "en"))
                                }}
                                """,
                    },
                    {
                        'name': 'altName',
                        'sparql': """
                                PREFIX gn: <http://www.geonames.org/ontology#>
                                SELECT ?altName (LANG(?altName) AS ?lang)
                                WHERE {{
                                    <{subject}> gn:alternateName ?altName
                                    FILTER (LANGMATCHES(LANG(?altName), "de") || LANGMATCHES(LANG(?altName), "en"))
                                }}
                                """,
                    },
                    {
                        'name': 'kind',
                        'sparql': """
                                PREFIX gn: <http://www.geonames.org/ontology#>
                                SELECT ?kind
                                WHERE {{
                                    <{subject}> gn:featureCode ?kind
                                }}
                                """,
                    },
                    {'name': 'lat-long',
                        'sparql': """
                                PREFIX gn: <http://www.geonames.org/ontology#>
                                PREFIX wgs84_pos: <http://www.w3.org/2003/01/geo/wgs84_pos#>
                                SELECT ?lat ?long
                                WHERE {{
                                    <{subject}> wgs84_pos:lat ?lat;
                                                wgs84_pos:long ?long
                                }}
                                """,

                    },
                    {'name': 'parent',
                     'sparql': """
                            PREFIX gn: <http://www.geonames.org/ontology#>
                            SELECT ?parent
                            WHERE {{
                                <{subject}> gn:parentCountry ?parent
                            }}
                            """,

                     },
                ]
            },
            {
                'base_url': 'http://d-nb.info/gnd/',
                'url_appendix': 'about/rdf',
                'attributes': [
                    {
                        'name': 'name',
                        'identifiers': (
                            (('objects', 'prefName', gndo + 'preferredNameForThePlaceOrGeographicName', None),),
                        )
                    },
                    {
                        'name': 'label',
                        'identifiers': (
                            (
                                ('objects', 'label', gndo + 'variantNameForThePlaceOrGeographicName', None),
                            ),
                        ),
                    },
                    {
                        'name': 'latlong',
                        'identifiers': (
                            (('objects', 'latlong_base', geo + 'hasGeometry', None), '>',
                             ('objects', 'latlong_geonode', geo + 'asWKT', None)
                             ),
                        )
                    },
                ]
            }
        ],
        'matching': {
            'attributes': {
                'name': {
                    'string': '{prefName}',
                    'regex': None
                },
                'kind': [{
                    'string': '{kind}',
                    'accessor': 'name',
                    'regex': None
                    },],
                'lat': {
                    'identifier': 'lat-long',
                    'string': '{lat}'
                },
                'lng': {
                    'identifier': 'lat-long',
                    'string': '{long}'
                }
            },
            'labels': {
                'alternative name': (
                    ('label', None),
                ),
                },
            'linked objects':
            [
            {
                'type': 'Place',
                'kind': 'located in',
                'object': 'parent'
            },
            ],
        }
    },
    'Person': {
        'data': [
            {
                'base_url': 'http://d-nb.info/gnd/',
                'url_appendix': 'about/rdf',
                'attributes': [
                    {
                        'name': 'name',
                        'identifiers': (
                            (
                             ('objects', 'prefNameNode', gndo + 'preferredNameEntityForThePerson', None),'>',
                             ('objects', 'forename', gndo + 'forename', None), '=',
                             ('objects', 'surname', gndo + 'surname', None)
                        ),
                            (
                              ('objects', 'prefNameNode', gndo + 'preferredNameEntityForThePerson', None), '>',
                                ('objects', 'descriptionNode', gndo + 'Description', None), '>',
                                ('objects', 'personalNameAddition', gndo + 'nameAddition', None), '=',
                                ('objects', 'personalNameCounting', gndo + 'counting', None), '=',
                                ('objects', 'personalName', gndo + 'personalName', None),
                            ),

                        )
                    },
                    {
                        'name': 'label',
                        'identifiers': (
                            (

                                ('objects', 'label', gndo + 'variantNameForThePerson', None),

                            ),
                        ),
                    },
                    {
                        'name': 'place of birth',
                        'identifiers': (
                        (
                            ('objects', 'place of birth', gndo + 'placeOfBirth', None),
                        ),
                    ),
                    },
                    {
                        'name': 'place of death',
                        'identifiers': (
                        (
                            ('objects', 'place of death', gndo + 'placeOfDeath', None),
                        ),
                    ),
                    },
                    {
                        'name': 'date of birth',
                        'identifiers': (
                        (
                            ('objects', 'date of birth', gndo + 'dateOfBirth', None),
                        ),
                    ),
                    },
                    {
                        'name': 'date of death',
                        'identifiers': (
                        (
                            ('objects', 'date of death', gndo + 'dateOfDeath', None),
                        ),
                    ),
                    },
                ]
            },
            {
                'base_url': 'https://www.wikidata.org/wiki/Special:EntityData/',
                'url_appendix': '.rdf',
                'attributes': [
                    {
                        'name': 'name',
                        'identifiers': (
                            (
                                ('objects', '')
                            )
                        )
                    },
                ]
            }
        ],
        'matching': {
            'attributes': {
                'name': "{self}",
                'first_name': (
                    (('forename', None),),
                    (('personalName', None),
                    ' ',
                     ('personalNameCounting', None)
                    ),
                ),
                'start_date_written': (
                    (('date of birth', None),),
                ),
                'end_date_written': (
                    (('date of death', None),),
                ),
            },
            'labels': {
                'alternative name': (
                    ('label', None),
                ),
            },
            'linked objects':
            [
            {
                'type': 'Place',
                'kind': 'born in',
                'object': (
                    ('place of birth', None),
            )
            },
            {
                'type': 'Place',
                'kind': 'died in',
                'object': (
                    ('place of death', None),
            )
            },
            ]
        }
    },
    'Event': {
        'data': [
            {
                'base_url': 'http://d-nb.info/gnd/',
                'url_appendix': 'about/rdf',
                'attributes': [
                    {
                        'name': 'name',
                        'identifiers': (
                            (('objects', 'prefName', gndo + 'preferredNameForTheSubjectHeading', None),),
                        )
                    },
                    {
                        'name': 'start',
                        'identifiers': (
                            (('objects', 'start_date', gndo + 'dateOfEstablishment', None),),
                        )
                    },
                    {
                        'name': 'end',
                        'identifiers': (
                            (('objects', 'end_date', gndo + 'dateOfTermination', None),),
                        )
                    },
                    {
                        'name': 'label',
                        'identifiers': (
                            (('objects', 'label', gndo + 'variantNameForTheSubjectHeading', None),),
                        )
                    },
                    {
                        'name': 'place of event',
                        'identifiers': (
                            (('objects', 'place of event', gndo + 'place', None),),
                        )
                    }
                ],
            }
        ],
        'matching': {
            'attributes': {
                'name': (
                    (('prefName', None),),
                ),
                'start_date_written': (
                    (('start_date', None),),
                ),
                'end_date_written': (
                    (('end_date', None),),
                )
            },
            'labels': {
                'alternative name': (
                    ('label', None),
                )
            },
            'linked objects': [
            {
                'type': 'Place',
                'kind': 'place of event',
                'object': (
                    ('place of event', None),
            )
            },
            ],
        }
    },
    'Institution': {
        'data': [
            {
                'base_url': 'http://d-nb.info/gnd/',
                'url_appendix': 'about/rdf',
                'attributes': [
                    {
                        'name': 'name',
                        'identifiers': (
                            (('objects', 'prefName', gndo + 'preferredNameForTheCorporateBody', None),),
                            (('objects', 'prefName', gndo + 'variantNameForTheCorporateBody', None),),
                        )
                    },
                    {
                        'name': 'alternativeName',
                        'identifiers': (
                            (('objects', 'alternativeName', gndo + 'variantNameForTheCorporateBody', None),),
                        )
                    },
                    {
                        'name': 'start',
                        'identifiers': (
                            (('objects', 'start', gndo + 'dateOfEstablishment', None),),
                        )
                    },
                    {
                        'name': 'end',
                        'identifiers': (
                            (('objects', 'end', gndo + 'dateOfTermination', None),),
                        )
                    },
                    {
                        'name': 'placeOfBusiness',
                        'identifiers': (
                            (('objects', 'placeOfBusiness', gndo + 'placeOfBusiness', None),),
                        )
                    },
                    {
                        'name': 'succeeding',
                        'identifiers': (
                            (('objects', 'succeeding', gndo + 'succeedingCorporateBody', None),),
                        )
                    },
                    {
                        'name': 'preceding',
                        'identifiers': (
                            (('objects', 'preceding', gndo + 'precedingCorporateBody', None),),
                        )
                    }
                ]
            }
        ],
        'matching': {
            'attributes': {
                'name': "{prefName}",
                'start_date_written': (
                    (('start', None),),
                ),
                'end_date_written': (
                    (('end', None),),
                )
            },
            'labels': {
                'alternative name': (
                    ('alternativeName', None),
                )
            },
            'linked objects': [
            {
                'type': 'Place',
                'kind': 'located in',
                'object': (
                    ('placeOfBusiness', None),
            )
            },
            {
                'type': 'Institution',
                'kind': 'preceding',
                'object': (
                    ('preceding', None),
            )
            },
            {
                'type': 'Institution',
                'kind': 'succeeding',
                'object': (
                    ('succeeding', None),
            )
            },
            ]
        }
    }
}


# Generated by Django 2.2.13 on 2020-07-22 12:31

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('apis_entities', '0003_auto_20200216_1152'),
    ]

    operations = [
        migrations.AlterField(
            model_name='person',
            name='gender',
            field=models.CharField(blank=True, choices=[('female', 'female'), ('male', 'male'), ('third gender', 'third gender')], max_length=15, null=True),
        ),
    ]



# Generated by Django 2.1.12 on 2020-01-21 12:27

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('apis_metainfo', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='Event',
            fields=[
                ('tempentityclass_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='apis_metainfo.TempEntityClass')),
            ],
            options={
                'abstract': False,
            },
            bases=('apis_metainfo.tempentityclass',),
        ),
        migrations.CreateModel(
            name='Institution',
            fields=[
                ('tempentityclass_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='apis_metainfo.TempEntityClass')),
            ],
            options={
                'abstract': False,
            },
            bases=('apis_metainfo.tempentityclass',),
        ),
        migrations.CreateModel(
            name='Person',
            fields=[
                ('tempentityclass_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='apis_metainfo.TempEntityClass')),
                ('first_name', models.CharField(blank=True, help_text='The persons´s forename. In case of more then one name...', max_length=255, null=True)),
                ('gender', models.CharField(blank=True, choices=[('female', 'female'), ('male', 'male')], max_length=15)),
            ],
            options={
                'abstract': False,
            },
            bases=('apis_metainfo.tempentityclass',),
        ),
        migrations.CreateModel(
            name='Place',
            fields=[
                ('tempentityclass_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='apis_metainfo.TempEntityClass')),
                ('lat', models.FloatField(blank=True, null=True, verbose_name='latitude')),
                ('lng', models.FloatField(blank=True, null=True, verbose_name='longitude')),
            ],
            options={
                'abstract': False,
            },
            bases=('apis_metainfo.tempentityclass',),
        ),
        migrations.CreateModel(
            name='Work',
            fields=[
                ('tempentityclass_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='apis_metainfo.TempEntityClass')),
            ],
            options={
                'abstract': False,
            },
            bases=('apis_metainfo.tempentityclass',),
        ),
    ]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['apis_core',
 'apis_core.apis_entities',
 'apis_core.apis_entities.api_mappings',
 'apis_core.apis_entities.management.commands',
 'apis_core.apis_entities.migrations',
 'apis_core.apis_entities.templatetags',
 'apis_core.apis_labels',
 'apis_core.apis_labels.migrations',
 'apis_core.apis_metainfo',
 'apis_core.apis_metainfo.migrations',
 'apis_core.apis_relations',
 'apis_core.apis_relations.migrations',
 'apis_core.apis_relations.templatetags',
 'apis_core.apis_tei',
 'apis_core.apis_vis',
 'apis_core.apis_vocabularies',
 'apis_core.apis_vocabularies.management.commands',
 'apis_core.apis_vocabularies.migrations',
 'apis_core.context_processors',
 'apis_core.default_settings',
 'apis_core.helper_functions',
 'browsing',
 'browsing.management.commands',
 'browsing.migrations',
 'browsing.templatetags',
 'infos',
 'infos.migrations',
 'infos.templatetags']

package_data = \
{'': ['*'],
 'apis_core': ['.vscode/*',
               'templates/apis_templates/*',
               'templates/apis_templates/autocomplete/*',
               'templates/apis_templates/registration/*'],
 'apis_core.apis_entities': ['fixtures/*',
                             'static/apis_entities/libraries/*',
                             'templates/apis_entities/*',
                             'templates/apis_entities/detail_views/*'],
 'apis_core.apis_labels': ['templates/apis_labels/*'],
 'apis_core.apis_metainfo': ['templates/apis_metainfo/*'],
 'apis_core.apis_relations': ['templates/apis_relations/*'],
 'apis_core.apis_vis': ['templates/apis_vis/*'],
 'browsing': ['templates/browsing/*', 'templates/browsing/tags/*'],
 'infos': ['templates/infos/*']}

install_requires = \
['Django>=3.1.4,<3.2.0',
 'PyYAML>=5.3.1,<6.0.0',
 'SPARQLWrapper>=1.8.5,<2.0.0',
 'convertdate>=2.3.0,<3.0.0',
 'dj-database-url>=0.5.0,<0.6.0',
 'django-admin-csvexport>=1.9,<2.0',
 'django-autocomplete-light>=3.8.1,<4.0.0',
 'django-cors-headers>=3.5.0,<4.0.0',
 'django-crispy-forms>=1.10.0,<2.0.0',
 'django-crum>=0.7.9,<0.8.0',
 'django-filter>=2.4.0,<3.0.0',
 'django-gm2m>=1.1.1,<2.0.0',
 'django-guardian>=2.3.0,<3.0.0',
 'django-leaflet>=0.27.1,<0.28.0',
 'django-model-utils>=4.1.1,<5.0.0',
 'django-reversion>=3.0.8,<4.0.0',
 'django-summernote>=0.8.11,<0.9.0',
 'django-tables2>=2.3.3,<3.0.0',
 'djangorestframework-csv>=2.1.0,<3.0.0',
 'djangorestframework-jsonschema>=0.1.1,<0.2.0',
 'djangorestframework-xml>=2.0.0,<3.0.0',
 'djangorestframework>=3.12.2,<4.0.0',
 'drf-spectacular==0.11.1',
 'jmespath>=0.10.0,<0.11.0',
 'jsonschema>=3.2.0,<4.0.0',
 'lxml>=4.6.2,<5.0.0',
 'mysqlclient>=2.0.3,<3.0.0',
 'pandas>=0.25.3,<0.26.0',
 'rdflib>=5.0.0,<6.0.0',
 'regex>=2020.11.13,<2021.0.0',
 'requests>=2.25.0,<3.0.0',
 'sentry-sdk>=0.19.5,<0.20.0',
 'tablib>=3.0.0,<4.0.0',
 'unicodecsv>=0.14.1,<0.15.0']

setup_kwargs = {
    'name': 'apis-core',
    'version': '0.16.15',
    'description': 'Base package for the APIS framework',
    'long_description': None,
    'author': 'Matthias Schlögl',
    'author_email': 'matthias.schloegl@oeaw.ac.at',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<3.7',
}


setup(**setup_kwargs)



from django.conf.urls import url

from . import views, api_views

app_name = 'apis_vis'

urlpatterns = [
    url(r'^heatmap/$', views.HeatMapView.as_view(), name='heatmap_view'),
    url(r'^heatmap-data/$', views.get_heatmap_data, name='get_heatmap_data'),
    url(r'^avg-age-data/$', views.get_average_age_data, name='get_avg_age_data'),
    url(r'^avg-age/$', views.AvgAge.as_view(), name='avgage_view'),
    url(r'^avg-members-data/$', views.get_average_members_data, name='get_avg_members_data'),
    url(r'^avg-members/$', views.MembersAmountPerYear.as_view(), name='avg_members_view'),
    url(
        r'^(?P<relation>[a-z]+)/data/$',
        api_views.GetVisJson.as_view(),
        name='person-institution-data'
    ),
    url(r'^inst-range-data/$', views.get_inst_range_data, name='get_inst_range_data'),
    url(r'^inst-range/$', views.InstRange.as_view(), name='inst_range_view'),
]



import pandas as pd
from django.contrib.auth.decorators import login_required
from django.contrib.contenttypes.models import ContentType
from django.http import HttpResponse


@login_required
def dl_vocabs_as_csv(request, model_name):
    kwargs = {
        "app_label": "apis_vocabularies",
        "entity": model_name
    }
    model = ContentType.objects.get(
        app_label=kwargs.get("app_label"), model=kwargs.get("entity").lower()
    ).model_class()
    data = [[x.id, x.name, x.parent_class] for x in model.objects.all()]
    columns = ['id', 'name', 'parent_class']
    df = pd.DataFrame(data, columns=columns)
    filename = f"{kwargs.get('entity').lower()}.csv"
    # Create the HttpResponse object with the appropriate CSV header.
    response = HttpResponse(content_type='text/csv')
    response['Content-Disposition'] = f"attachment; filename='{filename}'"
    df.to_csv(response, index=False)

    return response

# Create your views here.



from xml.sax.saxutils import escape, unescape

import lxml.etree as ET

from django.conf import settings
from django.utils.text import slugify

from .partials import TEI_NSMAP, tei_gen_header

from apis_core.apis_metainfo.models import Text
from apis_core.apis_entities.models import Person, Place, Event, Work, Institution


from collections import defaultdict

import pprint

pp = pprint.PrettyPrinter(indent=4)


def custom_escape(somestring):
    un_escaped = unescape(somestring)
    return escape(un_escaped)


# When serializing texts, we need to be able to look up
# the entity in DB to check we have all the texts (not just the
# annotated ones) — so need to map string values to classes.
# REPLACE THIS WITH LIST OF TEXTS IN ent_dict!
entity_type_string_to_class = {
    "Person": Person,
    "Place": Place,
    "Event": Event,
    "Work": Work,
    "Institution": Institution,
}

relation_type_to_tei_tag = {
    "events": "EVENT",  # What to do with this?
    "institutions": "orgName",
    "persons": "persName",
    "places": "placeName",
}


class TeiEntCreator:
    def __init__(
        self, ent_dict, base_url="entity/", include_entity_tagged_texts=False
    ):
        """
        Entry point: called from apis_core/apis_entities/api_renderers.py
        
        ent_dict is dict representing the entity

        This class is initialised, then has the serialize_full_doc method called
        on it, which in turn calls all the other methods on this class (apparently).


        """
        self.nsmap = TEI_NSMAP
        self.project = "APIS"
        s_url = getattr(settings, "APIS_BASE_URI", "http://apis.info/")
        self.base_url = f"{s_url}{base_url}"
        self.ent_dict = ent_dict
        self.ent_name = custom_escape(ent_dict.get("name", "No name provided"))
        self.ent_type = ent_dict.get("entity_type")
        self.ent_apis_id = ent_dict.get("id")
        self.gen_tei_header = tei_gen_header
        self.include_entity_tagged_texts = include_entity_tagged_texts

    def relation_groups(self):
        ent_dict = self.ent_dict
        ent_types = list(ent_dict["relations"].keys())
        self_type = "{}s".format(ent_dict["entity_type"].lower())
        relations = []
        for x in ent_types:
            group = []
            for y in ent_dict["relations"][x]:
                ent_key = list(y.keys())[2]
                rel = {}
                rel["rel_type"] = x
                rel["rel_label"] = slugify(y["relation_type"]["label"])
                if "target" in y.keys():
                    key_1 = "target"
                else:
                    key_1 = "source"
                rel["target"] = y[key_1]["id"]
                rel["target_name"] = y[key_1]["name"]
                group.append(rel)
            if group:
                relations.append(group)

        return relations

    def relation_notes(self):
        notes = []
        for group in self.relation_groups():
            note = ET.Element("{http://www.tei-c.org/ns/1.0}note")
            note_type = group[0]["rel_type"]
            note.attrib["type"] = note_type
            for item in group:
                ptr = ET.Element("ptr")
                ptr.attrib["type"] = item["rel_label"]
                ptr.attrib["target"] = "{}{}".format(self.base_url, item["target"])
                note.append(ptr)
            notes.append(note)
        return notes

    def uris_to_idnos(self):
        uris = []
        if self.ent_dict.get("uris"):
            for x in self.ent_dict.get("uris"):
                idno = ET.Element("idno")
                idno.attrib["type"] = "URL"
                if "d-nb.info" in x.get("uri"):
                    idno.attrib["subtype"] = "GND"
                idno.text = x.get("uri")
                uris.append(idno)
        return uris

    def create_work_node(self):
        work = ET.Element("{http://www.tei-c.org/ns/1.0}item")
        work.attrib["{http://www.w3.org/XML/1998/namespace}id"] = "work__{}".format(
            self.ent_apis_id
        )
        work.text = custom_escape(self.ent_dict.get("name"))
        for x in self.relation_notes():
            work.append(x)

        return work

    def create_event_node(self):
        event = ET.Element("{http://www.tei-c.org/ns/1.0}event")
        event.attrib["{http://www.w3.org/XML/1998/namespace}id"] = "event__{}".format(
            self.ent_apis_id
        )
        if self.ent_dict.get("start_date"):
            event.attrib["notBefore"] = self.ent_dict.get("start_date")
        if self.ent_dict.get("end_date"):
            event.attrib["notAfter"] = self.ent_dict.get("end_date")
        for x in self.relation_notes():
            event.append(x)
        label = ET.Element("label")
        label.text = self.ent_name
        event.append(label)
        if self.ent_dict.get("uris"):
            for x in self.ent_dict.get("uris"):
                event.attrib["ref"] = x.get("uri")
        return event

    def create_org_node(self):
        org = ET.Element("{http://www.tei-c.org/ns/1.0}org")
        org.attrib["{http://www.w3.org/XML/1998/namespace}id"] = "org__{}".format(
            self.ent_apis_id
        )
        orgName = ET.Element("orgName")
        orgName.text = self.ent_dict.get("name")
        if self.ent_dict.get("labels"):
            for x in self.ent_dict.get("labels"):
                node = ET.Element("orgName")
                node.attrib["type"] = "alt"
                try:
                    node_subtype = f'{x["label_type"]["name"]}'
                except KeyError:
                    node_subtype = False
                if node_subtype:
                    node.attrib["subtype"] = slugify(node_subtype)
                try:
                    node.attrib[
                        "{http://www.w3.org/XML/1998/namespace}lang"
                    ] = "{}".format(x["isoCode_639_3"])
                except KeyError:
                    pass
                node.text = x["label"]
                org.append(node)
        if self.ent_dict.get("start_date"):
            orgName.attrib["notBefore"] = self.ent_dict.get("start_date")
        if self.ent_dict.get("end_date"):
            orgName.attrib["notAfter"] = self.ent_dict.get("end_date")
        org.append(orgName)
        if self.uris_to_idnos():
            for x in self.uris_to_idnos():
                org.append(x)
        for x in self.relation_notes():
            org.append(x)
        return org

    def create_place_node(self):
        place = ET.Element("{http://www.tei-c.org/ns/1.0}place")
        place.attrib["{http://www.w3.org/XML/1998/namespace}id"] = "place__{}".format(
            self.ent_apis_id
        )
        placeName = ET.Element("placeName")
        placeName.text = self.ent_dict.get("name")
        if self.ent_dict.get("start_date"):
            placeName.attrib["notBefore"] = self.ent_dict.get("start_date")
        if self.ent_dict.get("end_date"):
            placeName.attrib["notAfter"] = self.ent_dict.get("end_date")
        place.append(placeName)
        if self.ent_dict.get("labels"):
            for x in self.ent_dict.get("labels"):
                node = ET.Element("placeName")
                node.attrib["type"] = "alt"
                try:
                    node_subtype = f'{x["label_type"]["name"]}'
                except KeyError:
                    node_subtype = False
                if node_subtype:
                    node.attrib["subtype"] = slugify(node_subtype)
                try:
                    node.attrib[
                        "{http://www.w3.org/XML/1998/namespace}lang"
                    ] = "{}".format(x["isoCode_639_3"])
                except KeyError:
                    pass
                node.text = x["label"]
                place.append(node)
        if self.ent_dict.get("lat"):
            coords = "{} {}".format(self.ent_dict["lat"], self.ent_dict["lng"])
            location = ET.Element("location")
            geo = ET.Element("geo")
            geo.text = coords
            location.append(geo)
            place.append(location)
        for x in self.relation_notes():
            place.append(x)
        if self.uris_to_idnos():
            for x in self.uris_to_idnos():
                place.append(x)
        return place

    def create_person_node(self):
        person = ET.Element("{http://www.tei-c.org/ns/1.0}person")
        person.attrib["{http://www.w3.org/XML/1998/namespace}id"] = "person__{}".format(
            self.ent_apis_id
        )
        persName = ET.Element("persName")
        surname = ET.Element("surname")
        surname.text = self.ent_dict.get("name")
        persName.append(surname)
        if self.ent_dict.get("first_name"):
            forename = ET.Element("forename")
            forename.text = self.ent_dict.get("first_name")
            persName.append(forename)
        person.append(persName)
        if self.ent_dict.get("labels"):
            for x in self.ent_dict.get("labels"):
                node = ET.Element("persName")
                node.attrib["type"] = "alt"
                try:
                    node_subtype = f'{x["label_type"]["name"]}'
                except KeyError:
                    node_subtype = False
                if node_subtype:
                    node.attrib["subtype"] = slugify(node_subtype)
                try:
                    node.attrib[
                        "{http://www.w3.org/XML/1998/namespace}lang"
                    ] = "{}".format(x["isoCode_639_3"])
                except KeyError:
                    pass
                node.text = x["label"]
                person.append(node)
        if self.ent_dict.get("start_date"):
            birth = ET.Element("birth")
            birth.attrib["when"] = self.ent_dict.get("start_date")
            birth.text = self.ent_dict.get("start_date_written")
            # If we wanted to add birth place, we'd need a specific lookup on
            # the place rel types here ... relations["places"]["relation_type"]["label"] == "BirthPlace"

            # It seems the only way to do this is to specify something in a controlled vocabulary --- some form of special status;

            if self.ent_dict.get("relations") and self.ent_dict.get("relations").get(
                "places"
            ):
                places = self.ent_dict.get("relations").get("places")
                pass

            person.append(birth)
        if self.ent_dict.get("end_date"):
            death = ET.Element("death")
            death.attrib["when"] = self.ent_dict.get("end_date")
            death.text = self.ent_dict.get("end_date_written")
            person.append(death)
        if self.ent_dict.get("profession"):
            for x in self.ent_dict.get("profession"):
                node = ET.Element("occupation")
                node.attrib["scheme"] = "#apis_vocabularies/professiontype/"
                node.attrib["code"] = "#{}".format(x["id"])
                node.text = x["name"]
                person.append(node)
        for x in self.relation_notes():
            person.append(x)
        if self.uris_to_idnos():
            for x in self.uris_to_idnos():
                person.append(x)
        return person

    def populate_header(self):
        main = f"{self.ent_type}: {self.ent_name}"
        header = self.gen_tei_header.format(main, self.project)
        return header

    def create_header_node(self):
        header = ET.fromstring(self.populate_header())
        return header

    def create_full_doc(self):
        ent_list = []
        doc = self.create_header_node()
        item = None
        if self.ent_type == "Person":
            item = self.create_person_node()
            ent_list = ET.Element("listPerson")
        elif self.ent_type == "Place":
            item = self.create_place_node()
            ent_list = ET.Element("listPlace")
        elif self.ent_type == "Institution":
            item = self.create_org_node()
            ent_list = ET.Element("listOrg")
        elif self.ent_type == "Event":
            item = self.create_event_node()
            ent_list = ET.Element("listEvent")
        elif self.ent_type == "Work":
            item = self.create_work_node()
            ent_list = ET.Element("list")
        # xpath modified to add entity nodes to <text type="entity"/>
        body = doc.xpath('//tei:text[@type="entity"]/tei:body', namespaces=self.nsmap)[
            0
        ]
        try:
            body.append(ent_list)
        except TypeError:
            pass
        ent_list.append(item)

        # Insert annotated texts into doc, as a separate text under
        # //tei:group
        groups = doc.xpath("//tei:group", namespaces=self.nsmap)[0]
        for text in self.build_annotated_texts_objects():
            groups.append(text)

        return doc

    def build_annotated_texts_objects(self):
        # Get annotations grouped by text

        annotations_by_text_id = group_annotations_by_text(self.ent_dict)

        texts = []

        # Here, we need to look up the entity in the DB
        ####  IDEALLY, pre-do this in initial query and add to ent_dict
        # though, this is likely to be quite a lot of stuff when not necessary...
        this_entity = entity_type_string_to_class[self.ent_type].objects.get(
            pk=self.ent_apis_id
        )
        this_entity_texts = {t for t in this_entity.text.all() if t.text}

        print([text.id for text in this_entity_texts])

        # Then, check all the entity texts to see whether it's an annotated one;
        # otherwise, we'll need to include it here...
        for text in this_entity_texts:
            if str(text.id) not in annotations_by_text_id:
                wrapped_text = "<p>{}</p>".format(
                    "</p><p>".join(text.text.split("\n\n"))
                )
                # Embed the text in some TEI tags
                embedded_text = '<text xml:id="text__{}"><body>{}</body></text>'.format(
                    text.id, wrapped_text
                )
                texts.append(embedded_text)

        # This loop here iterates texts from self.ent_dict annotations and embeds
        # annotations as TEI tags
        for text_id, anns in annotations_by_text_id.items():

            # Get the text object from database via its ID
            ### (in future, this should be embedded in ent_dict to avoid further query??)
            text_obj = Text.objects.get(pk=text_id)
            text_obj_pk = text_obj.tempentityclass_set.first().pk

            # If we *don't* want to include other entity_tagged texts,
            # skip this text if the object to which the text is not the
            # object we're serializing
            if not self.include_entity_tagged_texts and text_obj_pk != self.ent_apis_id:
                continue

            # On the other hand, if there is a text attached to this entity,
            # we might want to serialize it even if there is no annotation mentioning
            # this entity... (TAKEN CARE OF ABOVE)

            # Convert each annotation range from string-offset field ('34-52') into start and end indexes, and convert into tuple
            # of (start, end, tag) to run through standoff_to_inline function
            # (tag arg is got through looking up annotation type in relation_type_to_tei_tag — customisable
            # lookup at the top of this file)
            annotations = [
                (
                    int(ann["string_offset"].split("-")[0]),
                    int(ann["string_offset"].split("-")[1]),
                    f'{relation_type_to_tei_tag[ann["annotation_type"]]} ref="{ann["target"]}"',
                )
                for ann in anns
            ]

            tagged_text = stand_off_to_inline(text_obj.text, annotations)

            # Assuming the text has two-line separated paragraphs, wrap them in <p> tags
            wrapped_text = "<p>{}</p>".format("</p><p>".join(tagged_text.split("\n\n")))

            # Embed the text in some TEI tags
            embedded_text = '<text xml:id="text__{}"><body>{}</body></text>'.format(
                text_obj.id, wrapped_text
            )
            texts.append(embedded_text)

        # For each text string, convert to ET element
        texts = [convert_to_etree(text_as_string) for text_as_string in texts]
        return texts

    def serialize_full_doc(self):
        return ET.tostring(self.create_full_doc(), pretty_print=True, encoding="UTF-8")

    def export_full_doc(self):
        file = "temp.xml"
        with open(file, "wb") as f:
            f.write(
                ET.tostring(self.create_full_doc(), pretty_print=True, encoding="UTF-8")
            )
        return file


def convert_to_etree(text_as_string):
    # Function here to catch any XML syntax errors
    try:
        return ET.fromstring(text_as_string)
    # If there's some problem, most likely overlapping hierarchies,
    # we should return something to serialize
    except ET.XMLSyntaxError:
        return ET.fromstring(
            '<text type="conversion-error"><!-- There was an error inserting annotations into this document, probably due to overlaps --></text>'
        )


def group_annotations_by_text(ent_dict):
    annotations_by_text = defaultdict(list)

    for rel_type_name, rel_type_list in ent_dict.get("relations").items():

        for rel in rel_type_list:

            if rel.get("annotation"):
                for ann in rel.get("annotation", []):
                    text_id = ann["text_url"].split("/")[
                        -2
                    ]  # parse the url to get the text id...
                    target = rel.get("target", {}).get("url", None)
                    # add the annotation to the text-group, also adding in a type for conversion later
                    annotations_by_text[text_id].append(
                        {**ann, "annotation_type": rel_type_name, "target": target}
                    )

    return annotations_by_text


def stand_off_to_inline(text, annot):
    """
    Function from: https://gist.github.com/emsrc/c7aba4506214a20505ed

    Convert stand-off annotation to inline annotation
    Parameters
    ----------
    text : str
        unannotated text
    annot: list of tuples
        stand-off annotation as tuples in the form (i, j, tag)
        where integers i and j are the start and end character offset,
        and tag is a string for the tag label and - optionally - attributes.
    Returns
    -------
    inline: str
        text in xml format with inline annotation
    Notes
    -----
    Will not detect partly overlapping text spans, which will give rise
    to ill-formed xml.
    """
    # sort on decreasing span size
    annot.sort(key=lambda t: t[1] - t[0], reverse=True)

    # create dict mapping offsets to tags
    offsets2tags = defaultdict(list)

    for start, end, tag in annot:
        offsets2tags[start].append("<{}>".format(tag))
        offsets2tags[end].insert(0, "</{}>".format(tag.split()[0]))

    # merge text and tags
    parts = []
    i = None

    for j in sorted(offsets2tags.keys()) + [None]:
        parts.append(text[i:j])
        tags = "".join(offsets2tags[j])
        parts.append(tags)
        i = j

    return "".join(parts)




from django.conf import settings


def access_for_all(self, viewtype="list"):
    if self.request.user.is_authenticated:
        return self.request.user.is_authenticated
    elif viewtype == "list":
        try:
            access = settings.APIS_LIST_VIEWS_ALLOWED
        except AttributeError:
            access = False
            return access
        return access
    else:
        try:
            access = settings.APIS_DETAIL_VIEWS_ALLOWED
        except AttributeError:
            access = False
            return access
        return access


def access_for_all_function(user):
    if user.is_anonymous:
        print(getattr(settings, 'APIS_DETAIL_VIEWS_ALLOWED', False))
        return getattr(settings, 'APIS_DETAIL_VIEWS_ALLOWED', False)
    else:
        return True


ENTITIES_DEFAULT_COLS = [
    'start_date',
    'start_date_written',
    'end_date',
    'end_date_written',
    'text',
    'collection',
    'status',
    'source',
    'references',
    'notes',
]



import inspect
import re
import sys
import unicodedata
import yaml

from django.contrib.auth.models import User
from django.db import models
from django.utils.functional import cached_property
from reversion import revisions as reversion
from django.conf import settings


@reversion.register()
class VocabNames(models.Model):
    """List of Vocabulary names to allow the easy retrieval\
    of Vovcabulary names and classes from the VocabsBaseClass"""
    name = models.CharField(max_length=255)

    def get_vocab_label(self):
        return re.sub(r"([A-Z])", r" \1", self.name).strip()


@reversion.register()
class VocabsBaseClass(models.Model):
    """ An abstract base class for other classes which contain so called
    'controlled vocabulary' to describe subtypes of main temporalized
    entites"""
    choices_status = (
        ('rej', 'rejected'),
        ('ac', 'accepted'),
        ('can', 'candidate'),
        ('del', 'deleted')
    )
    name = models.CharField(max_length=255, verbose_name='Name')
    description = models.TextField(
        blank=True,
        help_text="Brief description of the used term.")
    parent_class = models.ForeignKey(
        'self', blank=True, null=True,
        on_delete=models.CASCADE
    )
    status = models.CharField(max_length=4, choices=choices_status, default='can')
    userAdded = models.ForeignKey(
        User, blank=True, null=True,
        on_delete=models.SET_NULL
    )
    vocab_name = models.ForeignKey(
        VocabNames, blank=True, null=True,
        on_delete=models.SET_NULL
    )

    def __str__(self):
        return self.label

    def save(self, *args, **kwargs):
        d, created = VocabNames.objects.get_or_create(name=type(self).__name__)
        self.vocab_name = d
        if self.name != unicodedata.normalize('NFC', self.name):  # secure correct unicode encoding
            self.name = unicodedata.normalize('NFC', self.name)
        super(VocabsBaseClass, self).save(*args, **kwargs)
        return self

    @cached_property
    def label(self):
        d = self
        res = self.name
        while d.parent_class:
            res = d.parent_class.name + ' >> ' + res
            d = d.parent_class
        return res


@reversion.register(follow=['vocabsbaseclass_ptr'])
class RelationBaseClass(VocabsBaseClass):
    """ An abstract base class for other classes which contain so called
    'controlled vocabulary' to describe the relations between main temporalized
    entities ('db_')"""

    name_reverse = models.CharField(
        max_length=255,
        verbose_name='Name reverse',
        help_text='Inverse relation like: "is sub-class of" vs. "is super-class of".',
        blank=True)

    def __str__(self):
        return self.name

    @cached_property
    def label_reverse(self):
        d = self
        if len(self.name_reverse) < 1:
            res = '(' + self.name + ')'
        else:
            res = self.name_reverse
        while d.parent_class:
            try:
                t = RelationBaseClass.objects.get(pk=d.parent_class.pk).name_reverse
                if len(t) < 1:
                    t = '(' + d.parent_class.name + ')'
            except Exception as e:
                t = '(' + d.parent_class.name + ')'
            res = t + ' >> ' + res
            d = d.parent_class
        return res

    def save(self, *args, **kwargs):
        if self.name_reverse != unicodedata.normalize('NFC', self.name_reverse):
            self.name_reverse = unicodedata.normalize('NFC', self.name_reverse)

        if self.name_reverse == "" or self.name_reverse == None:
            self.name_reverse = self.name + " [REVERSE]"

        super(RelationBaseClass, self).save(*args, **kwargs)
        return self


@reversion.register()
class VocabsUri(models.Model):
    """Class to store URIs for imported types. URI class from metainfo is not
    used in order to keep the vocabularies module/app seperated from the rest of the application.
    """
    uri = models.URLField()
    domain = models.CharField(max_length=255, blank=True)
    rdf_link = models.URLField(blank=True)
    vocab = models.ForeignKey(VocabsBaseClass, blank=True, null=True,
                              on_delete=models.CASCADE)
    # loaded: set to True when RDF was loaded and parsed into the data model
    loaded = models.BooleanField(default=False)
    # loaded_time: Timestamp when file was loaded and parsed
    loaded_time = models.DateTimeField(blank=True, null=True)

    def __str__(self):
        return self.uri

#######################################################################
#
#   entity types
#
#######################################################################


@reversion.register(follow=['vocabsbaseclass_ptr'])
class WorkType(VocabsBaseClass):
    """Holds controlled vocabularies about work-types"""
    pass


@reversion.register(follow=['vocabsbaseclass_ptr'])
class Title(VocabsBaseClass):
    """A person´s (academic) title"""
    abbreviation = models.CharField(max_length=10, blank=True)


@reversion.register(follow=['vocabsbaseclass_ptr'])
class ProfessionType(VocabsBaseClass):
    """Holds controlled vocabularies about profession-types"""
    pass


@reversion.register(follow=['vocabsbaseclass_ptr'])
class PlaceType(VocabsBaseClass):
    """Holds controlled vocabularies about place-types"""
    pass


@reversion.register(follow=['vocabsbaseclass_ptr'])
class InstitutionType(VocabsBaseClass):
    """Holds controlled vocabularies about institution-types"""
    pass


@reversion.register(follow=['vocabsbaseclass_ptr'])
class EventType(VocabsBaseClass):
    """Holds controlled vocabularies about event-types"""
    pass


@reversion.register(follow=['vocabsbaseclass_ptr'])
class LabelType(VocabsBaseClass):
    """Holds controlled vocabularies about label-types"""
    pass


@reversion.register(follow=['vocabsbaseclass_ptr'])
class CollectionType(VocabsBaseClass):
    """e.g. reseachCollection, importCollection """
    pass

@reversion.register(follow=['vocabsbaseclass_ptr'])
class TextType(VocabsBaseClass):
    """used to store the Text types for the forms"""
    entity = models.CharField(max_length=255)
    collections = models.ManyToManyField('apis_metainfo.Collection', blank=True)
    lang = models.CharField(
        max_length=3, blank=True, null=True,
        help_text="The ISO 639-3 (or 2) code for the label's language.",
        verbose_name='ISO Code', default='deu')



#######################################################################
#
#   relation types
#
#######################################################################



class AbstractRelationType(RelationBaseClass):
    """
    Abstract super class which encapsulates common logic between the different relationtypes and provides various methods
    relating to either all or a specific relationtypes.
    """

    class Meta:
        abstract = True

    _all_relationtype_classes = None
    _all_relationtype_names = None
    _related_entity_field_names = None


    # Methods dealing with all relationtypes
    ####################################################################################################################
    
    @classmethod
    def get_all_relationtype_classes(cls):
        """
        :return: list of all python classes of the relationtypes defined within this models' module  
        """

        if cls._all_relationtype_classes == None:

            relationtype_classes = []
            relationtype_names = []

            for relationtype_name, relationtype_class in inspect.getmembers(
                    sys.modules[__name__], inspect.isclass):

                if relationtype_class.__module__ == "apis_core.apis_vocabularies.models" and \
                        relationtype_name != "ent_class" and \
                        relationtype_name.endswith("Relation"):

                    relationtype_classes.append(relationtype_class)
                    relationtype_names.append(relationtype_name.lower())

            cls._all_relationtype_classes = relationtype_classes
            cls._all_relationtype_names = relationtype_names

        return cls._all_relationtype_classes


    @classmethod
    def get_relationtype_class_of_name(cls, relationtype_name):
        """
        :param entity_name: str : The name of an relationtype
        :return: The model class of the relationtype respective to the given name
        """

        for relationtype_class in cls.get_all_relationtype_classes():
            if relationtype_class.__name__.lower() == relationtype_name.lower():
                return relationtype_class

        raise Exception("Could not find relationtype class of name:", relationtype_name)


    @classmethod
    def get_all_relationtype_names(cls):
        """
        :return: list of all class names in lower case of the relationtypes defined within this models' module
        """

        if cls._all_relationtype_names == None:

            cls.get_all_relationtype_classes()

        return cls._all_relationtype_names


    # Methods dealing with related entities
    ####################################################################################################################
    
    @classmethod
    def get_related_entity_field_names(cls):
        """
        :return: a list of names of all ManyToMany field names relating to entities from the respective relationtype class

        E.g. for PersonPersonRelation.get_related_entity_field_names() or personpersonrelation_instance.get_related_entity_field_names() ->
        ['personB_set', 'personA_set']

        Note: this method depends on the 'generate_relation_fields' method in apis_entities.models which wires the ManyToMany Fields into the
        entities and respective relationtypes. It is nevertheless defined here within AbstractRelationType for documentational purpose.
        """

        if cls._related_entity_field_names == None:
            raise Exception("_related_entity_field_names was not initialized yet.")
        else:
            return cls._related_entity_field_names


    @classmethod
    def add_related_entity_field_name(cls, entity_field_name):
        """
        :param entity_field_name: the name of one of several ManyToMany fields created automatically
        :return: None

        Note: this method depends on the 'generate_relation_fields' method in apis_entities.models which wires the ManyToMany Fields into the
        entities and respective relationtypes. It is nevertheless defined here within AbstractRelationType for documentational purpose.
        """

        if cls._related_entity_field_names == None:
            cls._related_entity_field_names = []

        cls._related_entity_field_names.append(entity_field_name)


#######################################################################
# Person-Relation-Types
#######################################################################


@reversion.register(follow=['relationbaseclass_ptr'])
class PersonPersonRelation(AbstractRelationType):
    """Holds controlled vocabularies relation types of Persons and Persons"""
    pass


@reversion.register(follow=['relationbaseclass_ptr'])
class PersonPlaceRelation(AbstractRelationType):
    """Holds controlled vocabularies relation types of Persons and Places"""
    pass


@reversion.register(follow=['relationbaseclass_ptr'])
class PersonInstitutionRelation(AbstractRelationType):
    """Holds controlled vocabularies relation types of Persons and Persons"""
    pass


@reversion.register(follow=['relationbaseclass_ptr'])
class PersonEventRelation(AbstractRelationType):
    """Holds controlled vocabularies relation types of Persons and Events"""
    pass


@reversion.register(follow=['relationbaseclass_ptr'])
class PersonWorkRelation(AbstractRelationType):
    """Holds controlled vocabularies relation types of Persons and Works"""
    pass


#######################################################################
# Institution-Relation-Types
#######################################################################


@reversion.register(follow=['relationbaseclass_ptr'])
class InstitutionEventRelation(AbstractRelationType):
    """Holds controlled vocabularies relation types of Institutions and Events."""
    pass


@reversion.register(follow=['relationbaseclass_ptr'])
class InstitutionPlaceRelation(AbstractRelationType):
    """Holds controlled vocabularies relation types of Institutions and Places."""
    pass


@reversion.register(follow=['relationbaseclass_ptr'])
class InstitutionInstitutionRelation(AbstractRelationType):
    """Holds controlled vocabularies relation types of Institutions and Institutions."""
    pass


@reversion.register(follow=['relationbaseclass_ptr'])
class InstitutionWorkRelation(AbstractRelationType):
    """Holds controlled vocabularies relation types of Institutions and Works."""
    pass

#######################################################################
# Place-Relation-Types
#######################################################################


@reversion.register(follow=['relationbaseclass_ptr'])
class PlacePlaceRelation(AbstractRelationType):
    """Holds controlled vocabularies relation types of Places and Places"""
    pass


@reversion.register(follow=['relationbaseclass_ptr'])
class PlaceEventRelation(AbstractRelationType):
    """Holds controlled vocabularies relation types of Places and Events"""
    pass


@reversion.register(follow=['relationbaseclass_ptr'])
class PlaceWorkRelation(AbstractRelationType):
    """Holds controlled vocabularies relation types of Places and Works"""
    pass


#######################################################################
# Event-Relation-Types
#######################################################################


@reversion.register(follow=['relationbaseclass_ptr'])
class EventEventRelation(AbstractRelationType):
    """Holds controlled vocabularies relation types of Events and Events"""
    pass


@reversion.register(follow=['relationbaseclass_ptr'])
class EventWorkRelation(AbstractRelationType):
    """Holds controlled vocabularies relation types of Events and Works"""
    pass


#######################################################################
# Work-Relation-Types
#######################################################################


@reversion.register(follow=['relationbaseclass_ptr'])
class WorkWorkRelation(AbstractRelationType):
    """Holds controlled vocabularies relation types of Works and Works"""
    pass


a_ents = getattr(settings, 'APIS_ADDITIONAL_ENTITIES', False)

if a_ents:
    with open(a_ents, 'r') as ents_file:
        ents = yaml.load(ents_file, Loader=yaml.CLoader)
        print(ents)
        for ent in ents['entities']:
            for voc in ent.get('vocabs', []):
                attributes = {"__module__": __name__}
                ent_class = type(voc, (VocabsBaseClass,), attributes)
                globals()[ent['name']] = ent_class
            rels = ent.get("relations", [])
            base_ents = ['Person', 'Institution', 'Place', 'Work', 'Event']
            if isinstance(rels, str):
                if rels == 'all':
                    rels = base_ents + [x['name'].title() for x in ents['entities']]
            else:
                rels = base_ents + rels
            for r2 in rels:
                attributes = {"__module__":__name__}
                if r2 in base_ents:
                    rel_class_name = f"{r2.title()}{ent['name'].title()}"
                else:
                    rel_class_name = f"{ent['name'].title()}{r2.title()}"
                attributes = {"__module__": __name__}
                if f"{rel_class_name}Relation" not in globals().keys():
                    ent_class = type(f"{rel_class_name}Relation", (AbstractRelationType,), attributes)
                    globals()[f"{rel_class_name}Relation"] = ent_class




from django.test.runner import DiscoverRunner
from django.contrib.contenttypes.models import ContentType as ct
from django.db.models import fields as ft_dj
import random
import string


def create_data(ft, ln=5):
    print(ft)
    if ft == ft_dj.CharField or ft == ft_dj.TextField:
        return "".join(random.choices(string.ascii_letters + " ", k=ln))
    elif ft == ft_dj.IntegerField or ft == ft_dj.PositiveIntegerField:
        return int("".join(random.choices(string.digits, k=ln)))
    elif ft == ft_dj.DateField:
        return (
            f"{random.randint(1, 30)}.{random.randint(1, 12)}.{random.randint(1, 2020)}"
        )
    elif ft == ft_dj.BooleanField:
        return bool(random.getrandbits(1))
    else:
        print("didnt find any matching dt")


class APISTestRunner(DiscoverRunner):
    def setup_databases(self, **kwargs):
        old_names = super().setup_databases(**kwargs)

        lst_cont = [
            x.model_class()
            for x in ct.objects.filter(
                app_label__in=[
                    "apis_metainfo",
                    "apis_vocabularies",
                    "apis_entities",
                    "apis_relations",
                ]
            )
            .exclude(
                model__in=[
                    "tempentityclass",
                    "texttype_collections",
                    "vocabnames",
                    "vocabsuri",
                    "uricandidate",
                ]
            )
            .exclude(model__icontains="baseclass")
        ]
        lst_sort = [
            ct.objects.get(app_label="apis_metainfo", model="uri").model_class()
        ]
        len_lst_cont = len(lst_cont)
        len_lst_sort = len(lst_sort)

        while len(lst_cont) > len(lst_sort):
            for c in lst_cont:
                if c in lst_sort:
                    continue
                check = True
                for fld in c._meta.get_fields() + c._meta.many_to_many:
                    if "_set" in fld.name or fld.name in [
                        "parent_class",
                        "vocab_name",
                        "userAdded",
                        "groups_allowed",
                    ]:
                        continue
                    if fld.__class__.__name__ in ["ForeignKey", "ManyToManyField"]:
                        if fld.related_model not in lst_sort:
                            print(f"{c} / {fld}")
                            check = False
                if check:
                    lst_sort.append(c)
            if len_lst_cont == len(lst_cont) and len_lst_sort == len(lst_sort):
                print(lst_cont)
                print(lst_sort)
                print(len(lst_cont), len(lst_sort))
                fails = [x for x in lst_cont if x not in lst_sort]
                print(f"the following ents failed: {fails}")
                raise ValueError("no change in figures")
            len_lst_cont = len(lst_cont)
            len_lst_sort = len(lst_sort)
        dict_ents = {}
        l2 = lst_sort.pop()
        lst_sort.append(l2)
        print(lst_sort)
        for ent in lst_sort[1:]:
            print(ent)
            print(f"working on {ent}")
            c_1 = {}
            for fld in ent._meta.get_fields():
                print(f"working on field {fld.name}")
                ck = False
                for pn in ["_ptr", "set", "baseclass"]:
                    if pn in fld.name:
                        ck = True
                        break
                if fld.__class__.__name__ == "DateField" or ck:
                    continue
                elif fld.name in [
                    "parent_class",
                    "userAdded",
                    "vocab_name",
                    "id",
                    "vocabsuri",
                ]:
                    continue
                elif "date_written" in fld.name:
                    c_1[fld.name] = create_data(ft_dj.DateField)
                elif fld.__class__.__name__ == "ForeignKey":
                    c_1[fld.name] = dict_ents[fld.related_model.__name__][
                        random.randint(
                            0, len(dict_ents[fld.related_model.__name__]) - 1
                        )
                    ]
                elif (
                    fld.__class__.__name__ != "ManyToManyField"
                    and fld.__class__.__name__ != "ManyToOneRel"
                ):
                    c_1[fld.name] = create_data(
                        fld.__class__,
                        random.randint(
                            1,
                            (fld.max_length if fld.max_length else 3)
                            if hasattr(fld, "max_length")
                            else 3,
                        ),
                    )
            print(c_1)
            c_1_obj = ent.objects.create(**c_1)
            for fld in c._meta.many_to_many:
                if "_set" in fld.name:
                    continue
                getattr(c_1_obj, fld.name).add(
                    dict_ents[fld.related_model.__name__][
                        random.randint(
                            0, len(dict_ents[fld.related_model.__name__]) - 1
                        )
                    ]
                )
            if ent.__class__.__name__ not in dict_ents.keys():
                dict_ents[ent.__name__] = [
                    c_1_obj,
                ]
            else:
                dict_ents[ent.__name__].append(c_1_obj)
            print(f"created {c_1_obj}")
        return old_names


from functools import reduce
import copy
import importlib
import inspect

from django.conf import settings

# from django.contrib.contenttypes.models import ContentType
from django.urls import reverse
from rest_framework import pagination, serializers, viewsets
from rest_framework import renderers
from rest_framework.response import Response
from drf_spectacular.contrib.django_filters import (
    DjangoFilterBackend as DjangoFilterbackendSpectacular,
)
from drf_spectacular.utils import (
    extend_schema,
    extend_schema_field,
    OpenApiParameter,
    extend_schema_serializer,
)
from drf_spectacular.types import OpenApiTypes
from django import forms
from django_filters import rest_framework as filters
from .apis_metainfo.models import TempEntityClass
from .api_renderers import NetJsonRenderer
from .apis_relations.models import AbstractRelation
from apis_core.helper_functions.ContentType import GetContentTypes


if "apis_highlighter" in getattr(settings, "INSTALLED_APPS"):
    from apis_highlighter.highlighter import highlight_text_new
    from apis_highlighter.serializer import annotationSerializer
    from apis_highlighter.models import Annotation


try:
    MAX_AGE = settings.MAX_AGE
except AttributeError:
    MAX_AGE = 0


def deep_get(dictionary, keys, default=None):
    return reduce(
        lambda d, key: d.get(key, default) if isinstance(d, dict) else default,
        keys.split("."),
        dictionary,
    )


def create_query_parameters(entity):
    print(entity)
    for f in entity._meta.fields:
        print(f.name, f.__class__.__name__)


class CustomPagination(pagination.LimitOffsetPagination):
    def get_paginated_response(self, data):
        return Response(
            {
                "next": self.get_next_link(),
                "previous": self.get_previous_link(),
                "count": self.count,
                "limit": self.limit,
                "offset": self.offset,
                "results": data,
            }
        )


class ApisBaseSerializer(serializers.ModelSerializer):
    id = serializers.ReadOnlyField()
    label = serializers.SerializerMethodField(method_name="add_label")
    url = serializers.SerializerMethodField(method_name="add_uri")

    @extend_schema_field(OpenApiTypes.INT)
    def add_id(self, obj):
        return obj.pk

    @extend_schema_field(OpenApiTypes.STR)
    def add_label(self, obj):
        return str(obj)

    @extend_schema_field(OpenApiTypes.URI)
    def add_uri(self, obj):
        return self.context["view"].request.build_absolute_uri(
            reverse(
                "apis:apis_api:{}-detail".format(obj.__class__.__name__.lower()),
                kwargs={"pk": obj.pk},
            )
        )

    @extend_schema_field(OpenApiTypes.OBJECT)
    def add_type(self, obj):
        lst_type = ["kind", "type", "collection_type", "relation_type"]
        lst_kind = [
            x
            for x in obj._meta.fields
            if x.name in lst_type and "apis_vocabularies" in str(x.related_model)
        ]
        if len(lst_kind):
            pk_obj = getattr(obj, f"{lst_kind[0].name}_id")
            if pk_obj is not None:
                obj_type = getattr(obj, str(lst_kind[0].name))
                res = {
                    "id": pk_obj,
                    "url": self.context["view"].request.build_absolute_uri(
                        reverse(
                            f"apis:apis_api:{lst_kind[0].related_model.__name__.lower()}-detail",
                            kwargs={"pk": pk_obj},
                        )
                    ),
                    "type": obj_type.__class__.__name__,
                    "label": str(obj_type),
                    "parent_class": getattr(obj, "parent_class_id", None),
                }
                return res

    class Meta:
        model = TempEntityClass
        fields = ["id", "label", "url"]


class EntitySerializer(ApisBaseSerializer):
    type = serializers.SerializerMethodField(method_name="add_type")

    def add_type(self, instance):
        return instance.__class__.__name__

    class Meta(ApisBaseSerializer.Meta):
        fields = ApisBaseSerializer.Meta.fields + ["type"]


class LabelSerializer(ApisBaseSerializer):
    parent_id = serializers.PrimaryKeyRelatedField(
        many=False, source="parent_class_id", read_only=True
    )

    @extend_schema_field(OpenApiTypes.INT)
    def add_parent_id(self, obj):
        return obj.parent_class_id

    class Meta(ApisBaseSerializer.Meta):
        fields = ApisBaseSerializer.Meta.fields + [
            "parent_id",
        ]


class VocabsBaseSerializer(LabelSerializer, EntitySerializer):
    pass


class RelationObjectSerializer2(ApisBaseSerializer):
    relation_type = VocabsBaseSerializer(read_only=True)
    related_entity = serializers.SerializerMethodField(method_name="add_related_entity")

    @extend_schema_field(EntitySerializer)
    def add_related_entity(self, instance):
        for at in dir(instance):
            if (
                at.startswith("related_")
                and at.endswith("_id")
                and getattr(instance, at) != self._pk_instance
            ):
                return EntitySerializer(
                    getattr(instance, at[:-3]), context=self.context
                ).data

    class Meta(ApisBaseSerializer.Meta):
        fields = ApisBaseSerializer.Meta.fields + ["relation_type", "related_entity"]

    def __init__(self, *args, **kwargs):
        self._pk_instance = kwargs.pop("pk_instance")
        super(RelationObjectSerializer2, self).__init__(*args, **kwargs)


if "apis_highlighter" in getattr(settings, "INSTALLED_APPS"):

    class AnnotationSerializer(serializers.ModelSerializer):
        related_object = VocabsBaseSerializer(
            source="get_related_entity", read_only=True, many=False
        )

        class Meta:
            model = Annotation
            fields = ["id", "start", "end", "related_object"]


def generic_serializer_creation_factory():
    # lst_cont = [x.model_class() for x in ContentType.objects.filter(app_label__in=['apis_metainfo', 'apis_vocabularies', 'apis_entities', 'apis_relations']).exclude(model__in=["texttype_collections", "relationbaseclass"])]
    lst_cont = GetContentTypes().get_model_classes()
    not_allowed_filter_fields = [
        "useradded",
        "vocab_name",
        "parent_class",
        "vocab",
        "entity",
        "autofield",
    ]
    for cont in lst_cont:
        prefetch_rel = []
        select_related = []
        test_search = getattr(settings, cont.__module__.split(".")[1].upper(), False)
        entity_str = str(cont.__name__).replace(" ", "")
        entity = cont
        app_label = cont.__module__.split(".")[1].lower()
        exclude_lst = []
        if app_label == "apis_entities":
            exclude_lst = deep_get(test_search, "{}.api_exclude".format(entity_str), [])
        else:
            set_prem = getattr(settings, cont.__module__.split(".")[1].upper(), {})
            exclude_lst = deep_get(set_prem, "exclude", [])
            exclude_lst.extend(deep_get(set_prem, "{}.exclude".format(entity_str), []))
        exclude_lst_fin = [
            x for x in exclude_lst if x in [x.name for x in entity._meta.get_fields()]
        ]
        if entity_str.lower() == "text":
            exclude_lst_fin.extend(["kind", "source"])
        if app_label == "apis_relations":
            exclude_lst_fin.extend(["text", "collection"])
        for f in entity._meta.get_fields():
            if f.__class__.__name__ == "ManyToManyField":
                prefetch_rel.append(f.name)
            elif f.__class__.__name__ == "ForeignKey":
                select_related.append(f.name)

        class Meta:
            model = entity
            exclude = exclude_lst_fin

        def txt_serializer_add_text(self, instance):
            if self._inline_annotations:
                return self._txt_html
            else:
                return instance.text

        if "apis_highlighter" in getattr(settings, "INSTALLED_APPS"):

            @extend_schema_field(AnnotationSerializer(many=True))
            def txt_serializer_add_annotations(self, instance):
                if self._highlight:
                    return AnnotationSerializer(
                        self._annotations, context=self.context, many=True
                    ).data
                else:
                    return None

        def init_text_serializer(self, *args, **kwargs):
            super(self.__class__, self).__init__(*args, **kwargs)

            self._highlight = False
            self.fields["kind"] = LabelSerializer(many=False, read_only=True)

        def init_serializers(self, *args, **kwargs):
            super(self.__class__, self).__init__(*args, **kwargs)
            entity_str = self._entity.__name__
            app_label = self._app_label
            lst_labels_set = deep_get(
                getattr(settings, app_label.upper(), {}),
                "{}.labels".format(entity_str),
                [],
            )
            for f in self._entity._meta.get_fields():
                if getattr(settings, "APIS_API_EXCLUDE_SETS", False) and str(
                    f.name
                ).endswith("_set"):
                    if f.name in self.fields.keys():
                        self.fields.pop(f.name)
                    continue
                ck_many = f.__class__.__name__ == "ManyToManyField"
                if f.name in self._exclude_lst:
                    continue
                elif (
                    f.__class__.__name__
                    in [
                        "ManyToManyField",
                        "ForeignKey",
                    ]
                    and "apis_vocabularies" not in str(f.related_model)
                ):
                    self.fields[f.name] = ApisBaseSerializer(
                        many=ck_many, read_only=True
                    )
                elif f.__class__.__name__ in ["ManyToManyField", "ForeignKey"]:
                    self.fields[f.name] = LabelSerializer(many=ck_many, read_only=True)

        s_dict = {
            "id": serializers.ReadOnlyField(),
            "url": serializers.HyperlinkedIdentityField(
                view_name=f"apis:apis_api:{entity_str.lower()}-detail"
            ),
            "_entity": entity,
            "_exclude_lst": exclude_lst_fin,
            "_app_label": app_label,
            "Meta": Meta,
            "add_labels": lambda self, obj: {"id": obj.pk, "label": str(obj)},
            "__init__": init_serializers,
        }
        if entity_str.lower() == "text":
            s_dict["__init__"] = init_text_serializer

        s_dict_detail = copy.deepcopy(s_dict)

        serializer_class = type(
            f"{entity_str.title().replace(' ', '')}Serializer",
            (serializers.HyperlinkedModelSerializer,),
            s_dict,
        )

        def init_serializers_retrieve(self, *args, **kwargs):
            super(self.__class__, self).__init__(*args, **kwargs)
            entity_str = self._entity.__name__
            app_label = self._app_label
            lst_labels_set = deep_get(
                getattr(settings, app_label.upper(), {}),
                "{}.labels".format(entity_str),
                [],
            )
            for f in self._entity._meta.get_fields():
                if getattr(settings, "APIS_API_EXCLUDE_SETS", False) and str(
                    f.name
                ).endswith("_set"):
                    if f.name in self.fields.keys():
                        self.fields.pop(f.name)
                    continue
                ck_many = f.__class__.__name__ == "ManyToManyField"
                if f.name in self._exclude_lst:
                    continue
                elif (
                    f.__class__.__name__
                    in [
                        "ManyToManyField",
                        "ForeignKey",
                    ]
                    and "apis_vocabularies" not in str(f.related_model)
                ):
                    self.fields[f.name] = ApisBaseSerializer(
                        many=ck_many, read_only=True
                    )
                elif f.__class__.__name__ in ["ManyToManyField", "ForeignKey"]:
                    self.fields[f.name] = LabelSerializer(many=ck_many, read_only=True)
            # include = list(ContentType.objects.filter(app_label="apis_relations", model__icontains=entity_str).values_list('model', flat=True))
            include = [
                x
                for x in lst_cont
                if x.__module__ == "apis_core.apis_relations.models"
                and entity_str.lower() in x.__name__.lower()
            ]
            if len(include) > 0 and len(args) > 0:
                inst_pk2 = args[0].pk
                self.fields["relations"] = RelationObjectSerializer2(
                    read_only=True,
                    source="get_related_relation_instances",
                    many=True,
                    pk_instance=inst_pk2,
                )

        def init_text_serializer_retrieve(self, *args, **kwargs):
            super(self.__class__, self).__init__(*args, **kwargs)
            highlight = self.context.get("highlight", True)
            self._inline_annotations = False
            if highlight is not None and "apis_highlighter" in getattr(
                settings, "INSTALLED_APPS"
            ):
                self._highlight = highlight
                if self._highlight == "":
                    self._highlight = True
                if not isinstance(self._highlight, bool):
                    if self._highlight.lower() == "false":
                        self._highlight = False
                self._ann_proj_pk = self.context.get("ann_proj_pk", None)
                self._types = self.context.get("types", None)
                self._users_show = self.context.get("users_show", None)
                self._inline_annotations = self.context.get("inline_annotations", True)
                if not isinstance(self._inline_annotations, bool):
                    if self._inline_annotations.lower() == "false":
                        self._inline_annotations = False
                    elif self._inline_annotations.lower() == "true":
                        self._inline_annotations = True
                try:
                    self._txt_html, self._annotations = highlight_text_new(
                        self.instance,
                        set_ann_proj=self._ann_proj_pk,
                        types=self._types,
                        users_show=self._users_show,
                        inline_annotations=self._inline_annotations,
                    )
                    qs_an = {"text": self.instance}
                    if self._users_show is not None:
                        qs_an["users_added__in"] = self._users_show
                    if self._ann_proj_pk is not None:
                        qs_an["annotation_project_id"] = self._ann_proj_pk
                    #self._annotations = Annotation.objects.filter(
                    #    **qs_an
                    #)  # FIXME: Currently this QS is called twice (highlight_text_new)
                except Exception as e:
                    self._txt_html = ""
                    self._annotations = []
            else:
                self._highlight = False
            self.fields["kind"] = LabelSerializer(many=False, read_only=True)

        s_dict_detail["__init__"] = init_serializers_retrieve

        if entity_str.lower() == "text":
            s_dict_detail["__init__"] = init_text_serializer_retrieve
            s_dict_detail["txt_serializer_add_text"] = txt_serializer_add_text
            s_dict_detail["text"] = serializers.SerializerMethodField(
                method_name="txt_serializer_add_text"
            )
            if "apis_highlighter" in getattr(settings, "INSTALLED_APPS"):
                s_dict_detail[
                    "txt_serializer_add_annotations"
                ] = txt_serializer_add_annotations
                s_dict_detail["annotations"] = serializers.SerializerMethodField(
                    method_name="txt_serializer_add_annotations"
                )

        serializer_class_retrieve = type(
            f"{entity_str.title().replace(' ', '')}DetailSerializer",
            (serializers.HyperlinkedModelSerializer,),
            s_dict_detail,
        )

        allowed_fields_filter = {
            "IntegerField": ["in", "range", "exact"],
            "CharField": ["exact", "icontains", "iregex", "isnull"],
            "DateField": ["year", "lt", "gt", "year__lt", "year__gt", "exact"],
            "PositiveIntegerField": ["in", "range", "exact"],
            "AutoField": ["in", "exact"],
        }
        filterset_dict = {}
        filter_fields = {}

        for field in entity._meta.fields + entity._meta.many_to_many:
            if getattr(settings, "APIS_API_EXCLUDE_SETS", False) and "_set" in str(
                field.name.lower()
            ):
                continue
            if (
                field.name.lower() in not_allowed_filter_fields
                or field.name == "tempentityclass_ptr"
            ):
                continue
            elif field.__class__.__name__ in ["ForeignKey", "ManyToManyField"]:
                filter_fields[field.name] = ["exact"]
                if field.__class__.__name__ == "ForeignKey":
                    filter_fields[field.name].append("in")
                for f2 in field.related_model._meta.fields:
                    if f2.__class__.__name__ in [
                        "CharField",
                        "DateField",
                        "IntegerField",
                        "AutoField",
                    ]:
                        filter_fields[
                            f"{field.name}__{f2.name}"
                        ] = allowed_fields_filter[f2.__class__.__name__]
                continue
            if field.__class__.__name__ in allowed_fields_filter.keys():
                filter_fields[field.name] = allowed_fields_filter[
                    field.__class__.__name__
                ]
            else:
                filter_fields[field.name] = ["exact"]
        additional_filters = getattr(settings, "APIS_API_ADDITIONAL_FILTERS", False)
        if additional_filters:
            if entity_str in additional_filters.keys():
                for f1 in additional_filters[entity_str]:
                    if f1[0] not in filter_fields.keys():
                        filter_fields[f1[0]] = f1[1]

        class MetaFilter(object):
            model = entity
            fields = filter_fields

        filterset_dict["Meta"] = MetaFilter

        def get_queryset(self):
            if "apis_relations" in str(self.model):
                qs = self.model.objects.filter_for_user()
            else:
                qs = self.model.objects.all()
            if len(self._prefetch_rel) > 0:
                qs = qs.prefetch_related(*self._prefetch_rel)
            if len(self._select_related) > 0:
                qs = qs.select_related(*self._select_related)
            return qs

        @extend_schema(responses=serializer_class(many=True))
        def list_viewset(self, request):
            res = super(self.__class__, self).list(request)
            return res

        @extend_schema(
            parameters=[
                OpenApiParameter(
                    name="highlight",
                    description="Whether to add annotations or not, defaults to true",
                    type=OpenApiTypes.BOOL,
                ),
                OpenApiParameter(
                    name="inline_annotations",
                    description="Whether to add html5 mark tags for annotations to the text, defaults to false",
                    type=OpenApiTypes.BOOL,
                ),
                OpenApiParameter(
                    name="ann_proj_pk",
                    description="PK of the annotation project to use for annotations",
                    type=OpenApiTypes.INT,
                ),
                OpenApiParameter(
                    name="types",
                    description="Content type pks of annotation types to show. E.g. PersonPlace relations (comma sperated list)",
                    type=OpenApiTypes.STR,
                ),
                OpenApiParameter(
                    name="users_show",
                    description="Filter annotations for users. PKs of users, comma seperated list",
                    type=OpenApiTypes.STR,
                ),
            ],
            responses={200: serializer_class_retrieve},
        )
        def retrieve_view_txt(self, request, pk=None):
            res = super(self.__class__, self).retrieve(request, pk=pk)
            return res

        def get_serializer_context(self):
            context = super(self.__class__, self).get_serializer_context()
            if self.action == "retrieve" and self.model.__name__.lower() == "text":
                cont = {}
                cont["highlight"] = self.request.query_params.get("highlight", None)
                cont["ann_proj_pk"] = self.request.query_params.get("ann_proj_pk", None)
                cont["types"] = self.request.query_params.get("types", None)
                cont["users_show"] = self.request.query_params.get("users_show", None)
                cont["inline_annotations"] = self.request.query_params.get(
                    "inline_annotations", True
                )
                context.update(cont)
            return context

        def get_serializer_class_f(self, *arg, **kwargs):
            if self.action == "list":
                return self._serializer_class
            else:
                return self._serializer_class_retrieve

        viewset_dict = {
            "_select_related": select_related,
            "_prefetch_rel": prefetch_rel,
            "pagination_class": CustomPagination,
            "model": entity,
            "filter_backends": (DjangoFilterbackendSpectacular,),
            "filterset_fields": filter_fields,
            "depth": 2,
            "renderer_classes": (
                renderers.JSONRenderer,
                renderers.BrowsableAPIRenderer,
                NetJsonRenderer,
            ),
            "_serializer_class": serializer_class,
            "_serializer_class_retrieve": serializer_class_retrieve,
            "get_serializer_class": get_serializer_class_f,
            "get_serializer_context": get_serializer_context,
            "get_queryset": get_queryset,
            "dispatch": lambda self, request, *args, **kwargs: super(
                self.__class__, self
            ).dispatch(request, *args, **kwargs),
        }
        if entity_str.lower() == "text":
            viewset_dict["retrieve"] = retrieve_view_txt
        serializers_dict[serializer_class.__name__] = serializer_class
        views[f"{entity_str.lower().replace(' ', '')}"] = type(
            f"Generic{entity_str.title().replace(' ', '')}ViewSet",
            (viewsets.ModelViewSet,),
            viewset_dict,
        )


serializers_dict = dict()
views = dict()
filter_classes = dict()
lst_filter_classes_check = []
generic_serializer_creation_factory()



# generated by appcreator
from django import forms
from crispy_forms.helper import FormHelper
from crispy_forms.layout import Submit,  Layout, Fieldset, Div, MultiField, HTML
from crispy_forms.bootstrap import Accordion, AccordionGroup
from . models import AboutTheProject, TeamMember, ProjectInst


class ProjectInstFilterFormHelper(FormHelper):
    def __init__(self, *args, **kwargs):
        super(ProjectInstFilterFormHelper, self).__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.form_class = 'genericFilterForm'
        self.form_method = 'GET'
        self.helper.form_tag = False
        self.add_input(Submit('Filter', 'Search'))
        self.layout = Layout(
            Fieldset(
                'Basic search options',
                'id',
                'name',
                css_id="basic_search_fields"
                ),
            Accordion(
                AccordionGroup(
                    'More',
                    'website',
                    css_id="more"
                    ),
                )
            )


class ProjectInstForm(forms.ModelForm):
    class Meta:
        model = ProjectInst
        fields = "__all__"

    def __init__(self, *args, **kwargs):
        super(ProjectInstForm, self).__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.form_tag = True
        self.helper.form_class = 'form-horizontal'
        self.helper.label_class = 'col-md-3'
        self.helper.field_class = 'col-md-9'
        self.helper.add_input(Submit('submit', 'save'),)


class TeamMemberFilterFormHelper(FormHelper):
    def __init__(self, *args, **kwargs):
        super(TeamMemberFilterFormHelper, self).__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.form_class = 'genericFilterForm'
        self.form_method = 'GET'
        self.helper.form_tag = False
        self.add_input(Submit('Filter', 'Search'))
        self.layout = Layout(
            Fieldset(
                'Basic search options',
                'id',
                'name',
                css_id="basic_search_fields"
                ),
            Accordion(
                AccordionGroup(
                    'More',
                    'website',
                    'role',
                    css_id="more"
                    ),
                )
            )


class TeamMemberForm(forms.ModelForm):
    class Meta:
        model = TeamMember
        fields = "__all__"

    def __init__(self, *args, **kwargs):
        super(TeamMemberForm, self).__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.form_tag = True
        self.helper.form_class = 'form-horizontal'
        self.helper.label_class = 'col-md-3'
        self.helper.field_class = 'col-md-9'
        self.helper.add_input(Submit('submit', 'save'),)


class AboutTheProjectFilterFormHelper(FormHelper):
    def __init__(self, *args, **kwargs):
        super(AboutTheProjectFilterFormHelper, self).__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.form_class = 'genericFilterForm'
        self.form_method = 'GET'
        self.helper.form_tag = False
        self.add_input(Submit('Filter', 'Search'))
        self.layout = Layout(
            Fieldset(
                'Basic search options',
                'id',
                'description',
                css_id="basic_search_fields"
                ),
            Accordion(
                AccordionGroup(
                    'Authors',
                    'author',
                    css_id="more"
                    ),
                )
            )


class AboutTheProjectForm(forms.ModelForm):
    class Meta:
        model = AboutTheProject
        fields = "__all__"

    def __init__(self, *args, **kwargs):
        super(AboutTheProjectForm, self).__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.form_tag = True
        self.helper.form_class = 'form-horizontal'
        self.helper.label_class = 'col-md-3'
        self.helper.field_class = 'col-md-9'
        self.helper.add_input(Submit('submit', 'save'),)



from rest_framework import serializers

from .models import Collection, Source, TempEntityClass, Text, Uri


class CollectionSerializer(serializers.HyperlinkedModelSerializer):

    url = serializers.HyperlinkedIdentityField(
        view_name="apis:apis_api:text-detail", lookup_field="pk"
    )
    collection_type = serializers.HyperlinkedRelatedField(
        view_name="apis:apis_api:collectiontype-detail", lookup_field="pk", read_only=True
    )
    parent_class = serializers.HyperlinkedRelatedField(
        view_name="apis:apis_api:collection-detail", lookup_field="pk", read_only=True
    )

    class Meta:
        fields = ["url", "name", "description", "collection_type", "parent_class"]
        model = Collection


class TextSerializer(serializers.HyperlinkedModelSerializer):
    url = serializers.HyperlinkedIdentityField(
        view_name="apis:apis_api:text-detail", lookup_field="pk"
    )

    kind = serializers.HyperlinkedRelatedField(
        view_name="apis:apis_api:texttype-detail", lookup_field="pk", read_only=True
    )
    source = serializers.HyperlinkedRelatedField(
        view_name="apis:apis_api:source-detail", lookup_field="pk", read_only=True
    )

    class Meta:
        fields = ["kind", "text", "source", "url"]
        model = Text


class SourceSerializer(serializers.HyperlinkedModelSerializer):
    url = serializers.HyperlinkedIdentityField(
        view_name="apis:apis_api:source-detail", lookup_field="pk"
    )

    class Meta:
        fields = "__all__"
        model = Source


class UriSerializer(serializers.HyperlinkedModelSerializer):
    url = serializers.HyperlinkedIdentityField(
        view_name="apis:apis_api:uri-detail", lookup_field="pk"
    )
    entity = serializers.HyperlinkedRelatedField(
        view_name="apis:apis_api:tempentityclass-detail",
        lookup_field="pk",
        read_only=True,
    )

    class Meta:
        fields = "__all__"
        model = Uri


class TempEntityClassSerializer(serializers.HyperlinkedModelSerializer):
    url = serializers.HyperlinkedIdentityField(
        view_name="apis:apis_api:tempentityclass-detail", lookup_field="pk"
    )
    collection = serializers.HyperlinkedRelatedField(
        view_name="apis:apis_api:collection-detail", lookup_field="pk", read_only=True
    )
    text = serializers.HyperlinkedRelatedField(
        view_name="apis:apis_api:text-detail", lookup_field="pk", read_only=True
    )
    source = serializers.HyperlinkedRelatedField(
        view_name="apis:apis_api:source-detail", lookup_field="pk", read_only=True
    )

    class Meta:
        fields = "__all__"
        model = TempEntityClass
